import json
import os
from abc import abstractmethod
from collections import deque
from copy import deepcopy
from typing import Any, Dict, List, Optional

import torch

from mppq.data import convert_any_to_python_primary_type as _convert
from mppq.defs import SingletonMeta
from mppq.ir.base.opdef import Operation, Variable
from mppq.ir.base.quantize import QuantableOperation
from mppq.logger import error, warning
from mppq.quant import QuantizationStates, TargetPrecision
from mppq.storage import Serializable


class BaseGraph(Serializable):
    r"""Graph is a PPQ Internal Represtation Data Structure.

    计算图是一个有向图，其中节点对应于操作或变量。
    变量可以将其值输入到操作中，操作可以将其输出输入到其他操作中。这样，图中的每个节点都定义了变量的函数。

    输入节点和从节点输出的值称为张量，这只是一个多维数组的花哨说法。
    因此，它也包括标量、向量和矩阵以及更高秩的张量。

    PPQ 创建的计算图包含量化信息以及操作和变量。因此可以说它是一个专为量化设计的计算图。

    所有的量化相关信息都存储在图及其操作中。
    """

    def __init__(self, name: str) -> None:
        super().__init__()
        self.operations: Dict[str, Operation] = {}
        self.variables: Dict[str, Variable] = {}
        self.inputs: Dict[str, Variable] = {}
        self.outputs: Dict[str, Variable] = {}
        self.name = name
        self._num_of_generated_var = 0
        self._num_of_generated_op = 0
        self._detail: Dict[str, Any] = {}

    def parameters(self) -> List[torch.Tensor]:
        r"""获取所有参数变量"""
        parameters = []
        for var in self.variables.values():
            if var.is_parameter:
                parameters.append(var.value)
        return parameters

    def set_extension_attrib(self, attrib: str, value: Any):
        r"""设置扩展属性"""
        self._detail[attrib] = value

    @property
    def extension_attrib(self):
        r"""获取扩展属性"""
        return self._detail

    def append_operation(self, operation: Operation):
        r"""添加一个新算子到图中"""
        assert isinstance(operation, Operation)
        for var in operation.inputs + operation.outputs:
            if var.name not in self.variables:
                self.append_variable(var)
        if operation.name in self.operations:
            raise KeyError(f"Duplicated Operation({operation}) was found!")
        self.operations[operation.name] = operation

    def append_variable(self, var: Variable):
        r"""手动添加一个变量到图中。

        Note:

            变量的消费者必须提前添加或已存在。
        """
        assert isinstance(var, Variable)
        if not all([dest_op.name in self.operations for dest_op in var.dest_ops]):
            raise RuntimeError(
                f"Inserting Variable {var} has a related Operation(dest_op) "
                "which are not included in this graph yet, "
                "insert such Operations before inserting this."
            )
        if var.name in self.variables:
            raise KeyError(f"Duplicated Variable({var}) was found!")
        self.variables[var.name] = var

    def get_downstream_operations(self, operation: Operation) -> List[Operation]:
        r"""获取算子的后驱算子"""
        assert isinstance(operation, Operation)
        if operation.name not in self.operations:
            raise KeyError(f"Operation {operation.name} not in current graph.")
        downstream_ops = []
        for output_var in operation.outputs:
            downstream_ops.extend(output_var.dest_ops)
        return downstream_ops

    def get_upstream_operations(self, operation: Operation) -> List[Operation]:
        r"""获取算子的前驱算子"""
        assert isinstance(operation, Operation)
        if operation.name not in self.operations:
            raise KeyError(f"Operation {operation.name} not in current graph.")
        upstream_ops = []
        for input_var in operation.inputs:
            if input_var.source_op is not None:
                upstream_ops.append(input_var.source_op)
        return upstream_ops

    def topological_sort(self) -> List[Operation]:
        r"""返回拓扑排序后的算子列表"""
        visited = {operation.name: False for operation in self.operations.values()}
        sort_ret, pop_list = [], deque()
        num_of_inputs = {
            operation.name: len(self.get_upstream_operations(operation))
            for operation in self.operations.values()
        }

        # initialization
        for op_name, n_input in num_of_inputs.items():
            if n_input == 0:
                pop_list.append(op_name)

        # topological sort
        for _ in range(len(visited)):
            if len(pop_list) == 0:
                break
            op_name = pop_list.popleft()
            op = self.operations[op_name]
            for post_op in self.get_downstream_operations(op):
                num_of_inputs[post_op.name] -= 1
                if num_of_inputs[post_op.name] == 0:
                    pop_list.append(post_op.name)
            visited[op.name] = True
            sort_ret.append(op)
        if all(visited.values()):
            return sort_ret
        else:
            error("Some operation can not be sorted (might due to circular reference)")
            error("\n".join(op_name for op_name in visited if not visited[op_name]))
            raise RuntimeError("Topological Sort failed.")

    def insert_op_on_var(self, inserting_op: Operation, var: str):
        """Insert one operation to current graph. Inserting operation will
        replace var.dest_ops and automatically connect to inserting_op.

        Before insertion:
            op1 -> var -> op2

        After insertion:
            op1 -> var -> inserting_op -> link_var(generated) -> op2

        ATTENTION:

            Inserting operation must be an empty operation with no input and output
            variables linked to it.
        """
        if not isinstance(var, str):
            raise TypeError(
                f"Needs a variable name(str) here, however {type(var)} was given"
            )
        if var not in self.variables:
            raise KeyError(
                f"Can not inserting operation at variable {var}, variable not found."
            )
        if inserting_op.num_of_input != 0 or inserting_op.num_of_output != 0:
            raise PermissionError(
                "Can only insert operation with no input and output variables."
            )

        variable = self.variables[var]
        # add to graph.
        if inserting_op.name not in self.operations:
            self.append_operation(inserting_op)
        # create all links.
        link_var = self.create_variable(
            is_parameter=False,
            dest_ops=variable.dest_ops.copy(),
            source_op=inserting_op,
        )

        inserting_op.inputs.append(variable)
        inserting_op.outputs.append(link_var)

        variable.dest_ops.clear()
        variable.dest_ops.append(inserting_op)

        for op in link_var.dest_ops:
            op.inputs[op.inputs.index(variable)] = link_var

        if var in self.outputs:
            self.outputs.pop(var)
            self.outputs[link_var.name] = link_var

    def insert_op_before(self, op0: Operation, op1: Operation, input_idx: int = 0):
        """Insert an op just before given op. This function will insert given
        op A to variable B.inputs[input_idx]

        Args:
            op0 (Operation): Inserting Op, should has no input and output variable that
                links to it.
            op1 (Operation): before this op.
            input_idx (int, optional): For case that B has more than 1 input variable,
                user should use parameter input_idx to identify which variable is used.
        """
        if input_idx >= op1.num_of_input:
            raise ValueError("Input index out of range.")
        if op0.num_of_input != 0 or op0.num_of_output != 0:
            raise ValueError(
                "Can only insert op that has no input and output variable."
            )
        var = op1.inputs[input_idx]
        var.dest_ops[var.dest_ops.index(op1)] = op0
        op1.inputs[input_idx] = self.create_variable()
        op1.inputs[input_idx].source_op = op0
        op1.inputs[input_idx].dest_ops.append(op1)
        op0.inputs.append(var)
        op0.outputs.append(op1.inputs[input_idx])

    def insert_op_after(self, op0: Operation, op1: Operation, output_idx: int = 0):
        """Insert an op just after given op. This function will insert given op
        A to variable B.outputs[output_idx]

        Args:
            op0 (Operation): Inserting Op, should has no input and output variable that
                links to it.
            op1 (Operation): after this op.
            output_idx (int, optional): For case that B has more than 1 output variable,
                user should use parameter output_idx to identify which variable is used.
        """
        if output_idx >= op1.num_of_output:
            raise ValueError("Output index out of range.")
        if op0.num_of_input != 0 or op0.num_of_output != 0:
            raise ValueError(
                "Can only insert op that has no input and output variable."
            )
        var = op1.outputs[output_idx]
        var.source_op = op0
        op1.outputs[output_idx] = self.create_variable()
        op1.outputs[output_idx].source_op = op1
        op1.outputs[output_idx].dest_ops.append(op0)
        op0.outputs.append(var)
        op0.inputs.append(op1.outputs[output_idx])

    def create_link_with_op(
        self,
        op0: Optional[Operation],
        op1: Optional[Operation],
        variable: Optional[Variable] = None,
    ):
        """Create a link with given variable from upstream_op to downstream_op
        variable will be appended to upstream_op's output and downstream_op's
        input given variable must have empty source_op or its source_op == upstream_op.

        Sometime you may want to link a single upstream_op to many downstream_ops with
        a same variable, you are supposed to invoke this function for each downstream_op
        then.

        You can set upstream_op = None if your variable is a parameter variable.

        Example::

            create_link_with_op(var1, op1, op2)
            create_link_with_op(var1, op1, op3)

        Will makes:
                  --> op2
            op1 --|
                  --> op3
        """
        if variable is None:
            variable = self.create_variable()
        if variable.name not in self.variables:
            raise KeyError(
                f"Can not find your variable {variable.name} in current graph."
            )
        if op0 is not None and op0.name not in self.operations:
            raise KeyError(f"Can not find your operation {op0.name} in current graph.")
        if op1 is not None and op1.name not in self.operations:
            raise KeyError(f"Can not find your operation {op1.name} in current graph.")

        if variable.source_op is None and op0 is not None:
            variable.source_op = op0
        if variable.source_op != op0:
            raise PermissionError(
                f"Can not create link with variable {variable}, "
                f"cause its source operations != {op0}"
            )

        # For complex graph, following logic might have some error.
        if op0 is not None and variable not in op0.outputs:
            op0.outputs.append(variable)
        if op1 is None:
            return
        if op1 is not None and variable not in op1.inputs:
            variable.dest_ops.append(op1)
            op1.inputs.append(variable)
        else:
            variable.dest_ops.append(op1)
            op1.inputs.append(variable)
            warning(
                "You are trying to link variable with operation, "
                f"however Variable {variable.name} has already linked "
                f"with downstream op {op1.name}"
            )

    def create_link_with_var(self, op0: Variable, op1: Variable):
        """connect upstream_variable.source_op with downstream_variable.dest_ops,
        downstream variable will be eliminated by this function.

        downstream_variable must have None as its source_op.
        """
        if op0 is not None and op0.name not in self.variables:
            raise KeyError(f"Can not find your variable {op0.name} in current graph.")
        if op1 is not None and op1.name not in self.variables:
            raise KeyError(f"Can not find your variable {op1.name} in current graph.")

        if op1.source_op is not None:
            raise PermissionError(
                f"Can not create link with variable {op0.name} & {op1.name}, "
                "Cause downstream variable has a non-empty source op"
            )

        dest_ops = op1.dest_ops
        for dest_op in dest_ops:
            dest_op.inputs[dest_op.inputs.index(op1)] = op0
            op0.dest_ops.append(dest_op)
        op1.dest_ops.clear()
        self.remove_variable(op1)
        return self

    def remove_operation(
        self,
        removing_op: Operation,
        keep_coherence: bool = False,
        remove_unlinked_variable: bool = False,
    ):
        """Remove operation from graph, this function will unlink removing
        operation from current graph, pop it from graph.operations, and remove
        it from all its input and output variables.

        Parameters of this removing operations will be removed from graph by this
        function, without warning.

        Args:
            keep_coherence (bool): if keep_coherence = True, PPQ will link downstream
                operations of removing op to the upstream operation. If there is more
                than 1 input and output variable, PPQ will link input[0] with output[0]
        """
        if removing_op.name not in self.operations:
            raise KeyError(
                f"Can not remove operation {removing_op.name}, operation not found."
            )

        # removing all parameters first.
        for parameter in removing_op.inputs.copy():
            if keep_coherence and removing_op.type in {"Constant", "Identity"}:
                break
            if parameter.is_parameter:
                parameter.dest_ops.clear()
                # pylint: disable=protected-access
                parameter._value = None  # clear memory.
                removing_op.inputs.remove(parameter)

                self.variables.pop(parameter.name)

        related_vars = [var for var in removing_op.inputs + removing_op.outputs]
        input_var, output_var = (
            removing_op.inputs[0] if removing_op.num_of_input >= 1 else None,
            removing_op.outputs[0] if removing_op.num_of_output >= 1 else None,
        )

        # remove operation from its output variables
        for _output_var in removing_op.outputs:
            _output_var.source_op = None
        removing_op.outputs.clear()

        # remove operation from its input variables
        for _input_var in removing_op.inputs:
            if removing_op in _input_var.dest_ops:
                _input_var.dest_ops.remove(removing_op)
        removing_op.inputs.clear()

        if input_var is not None and output_var is not None and keep_coherence:

            removing_var = output_var
            dest_ops = removing_var.dest_ops
            is_graph_output = removing_var.name in self.outputs

            for op in dest_ops:
                op.inputs[op.inputs.index(removing_var)] = input_var
                input_var.dest_ops.append(op)
            removing_var.dest_ops.clear()
            removing_var.source_op = None
            self.remove_variable(removing_var)

            if is_graph_output:
                self.mark_variable_as_graph_output(input_var)

        self.operations.pop(removing_op.name)

        if remove_unlinked_variable:
            for var in related_vars:
                if (
                    var.source_op is None
                    and len(var.dest_ops) == 0
                    and var.name in self.variables
                ):
                    self.remove_variable(var)

        return self

    def remove_variable(self, removing_var: Variable):
        """Remove variable from graph, this function will unlink removing
        variable from current graph, pop it from graph.variables, and remove it
        from its source op and dest ops.
        """
        if removing_var.name not in self.variables:
            raise KeyError(
                f"Can not remove variable {removing_var.name}, variable not found."
            )

        # remove from source operation
        source_op = removing_var.source_op
        if source_op is not None:
            assert isinstance(source_op, Operation)
            if removing_var in source_op.outputs:
                source_op.outputs.remove(removing_var)
            removing_var.source_op = None

        # remove from all dest ops
        for dest_op in removing_var.dest_ops:
            assert isinstance(dest_op, Operation)
            if removing_var in dest_op.inputs:
                dest_op.inputs.remove(removing_var)
        removing_var.dest_ops.clear()

        if removing_var.name in self.outputs:
            self.outputs.pop(removing_var.name)

        if removing_var.name in self.inputs:
            self.inputs.pop(removing_var.name)

        self.variables.pop(removing_var.name)
        return self

    def create_operation(
        self,
        op_type: str,
        name: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        platform: TargetPrecision = TargetPrecision.UNSPECIFIED,
        inputs: Optional[List[Variable]] = None,
        outputs: Optional[List[Variable]] = None,
    ) -> Operation:
        """Create an operation and attach it it current graph. op_type is
        mandatory here, however op_name is not required. PPQ will automatically
        generates a name for your operation:
        PPQ_Operation_{self._num_of_generated_op}.

        Use this function carefully, cause once your network is quantized, simply
        create an operation via this function might cause unexpected error.
        Beawre that operation created by this function has no meta data and
        quantization info, which is needed to export and executing your graph.

        Do not set inputs and outputs via this function, to link your operation with
        others, use graph.create_link_with_var instead.
        """
        if inputs is None:
            inputs = []
        if outputs is None:
            outputs = []

        if name is None:
            name = f"PPQ_Operation_{self._num_of_generated_op}"
            self._num_of_generated_op += 1

        if not isinstance(inputs, list):
            raise TypeError(
                f"A list of input variable is required for creating operation, "
                f"however {type(inputs)} was given"
            )
        created = Operation(
            name=name,
            op_type=op_type,
            attributes=attributes or {},
            precision=platform,
            inputs=inputs,
            outputs=outputs,
        )
        self.append_operation(created)

        for item in inputs:
            if not isinstance(item, Variable):
                raise TypeError(
                    f"A list contains variables is required for creating operation, "
                    f"however there is a {type(item)} in your input list."
                )
            item.dest_ops.append(created)

        if not isinstance(outputs, list):
            raise TypeError(
                f"A list of output variable is required for creating operation, "
                f"however {type(inputs)} was given"
            )
        for item in outputs:
            if not isinstance(item, Variable):
                raise TypeError(
                    f"A list contains variables is required for creating operation, "
                    f"however there is a {type(item)} in your output list."
                )
            item.source_op = created
        return created

    def create_variable(
        self,
        name: Optional[str] = None,
        value: Optional[Any] = None,
        is_parameter: bool = False,
        dest_ops: Optional[List[Operation]] = None,
        source_op: Optional[Operation] = None,
    ) -> Variable:
        """Create a variable and attach it it current graph. PPQ will
        automatically generates a name for your variable:
        PPQ_Variable_{self._num_of_generated_op}.

        Use this function carefully, cause once your network is quantized, simply
        create an variable via this function might cause unexpected error.
        You'd better invoke this function before running your quantizer.

        If dest_ops and source_op is not None, this function will auto link created
        variable with them.
        """
        if name is None:
            name = f"PPQ_Variable_{self._num_of_generated_var}"
            self._num_of_generated_var += 1

        created = Variable(
            name=name,
            value=value,
            is_parameter=is_parameter,
            dest_ops=dest_ops,
            source_op=source_op,
        )

        if dest_ops is not None:
            for op in dest_ops:
                if not isinstance(op, Operation):
                    raise TypeError(
                        "Parameter dest ops should be a list of Operation, "
                        f"however {type(op)} was given."
                    )
                op.inputs.append(created)

        if source_op is not None:
            if not isinstance(source_op, Operation):
                raise TypeError(
                    "Parameter dest ops should be an Operation, "
                    f"however {type(source_op)} was given."
                )
            source_op.outputs.append(created)

        self.append_variable(created)
        return created

    def mark_variable_as_graph_input(self, var: Variable):
        r"""标记该变量为当前图的输入。

        Note:

            不能将输出标记为输入。
        """
        assert isinstance(var, Variable)
        var_name = var.name
        if var_name not in self.variables:
            raise KeyError(f"Can not find variable {var_name} within current graph.")
        if var_name in self.inputs:
            return
        if var_name in self.outputs:
            raise KeyError(
                f"Can not mark variable {var_name} as graph input, "
                "cause it is graph output."
            )
        self.inputs[var_name] = self.variables[var_name]

    def mark_variable_as_graph_output(self, var: Variable):
        r"""标记该变量为当前图的输出。

        Note:

            可以将输入标记为输出。
        """
        assert isinstance(var, Variable)
        var_name = var.name
        if var_name not in self.variables:
            raise KeyError(f"Can not find variable {var_name} within current graph.")
        if var_name in self.outputs:
            return
        self.outputs[var_name] = self.variables[var_name]

    def copy(self, copy_value: bool = False):  # noqa: C901
        """Clone current graph. Use parameter copy_value to control whether to
        do a Shallow Copy or Deep Copy.

        For copy_value = True, there will be a copy of each parameter in your network.
            ATTENTION: it might cause gpu memory overflow.
        For copy_value = False, cloned network will share the same parameter tensor of
            current one.

        ATTENTION: all quantization config will be cloned,
            all scales and offsets will be cloned even with copy_valye = False.

        Shallow Copy: Shallow repetition is quicker.
        However, it's “lazy” it handles pointers and references.
        Rather than creating a contemporary copy of the particular knowledge the
        pointer points to, it simply copies over the pointer price.
        So, each the first and therefore the copy can have pointers that reference
        constant underlying knowledge.

        Deep Copy: Deep repetition truly clones the underlying data.
        It is not shared between the first and therefore the copy.
        """

        cloned = BaseGraph(name=self.name)
        for op in self.operations.values():
            cloned.append_operation(deepcopy(op))
        for var in self.variables.values():
            cloned.append_variable(var.copy(copy_value=copy_value))

        # notice that all operations is copied without link, so do all variables
        # relink them with following code
        config_dict = {}
        for op in self.operations.values():
            assert (
                op.name in cloned.operations
            ), f"Graph Copy Error, Operation {op.name} is not correctly cloned"
            c_op = cloned.operations[op.name]
            for i_var in op.inputs:
                assert (
                    i_var.name in cloned.variables
                ), f"Graph Copy Error, Variable {i_var.name} is not correctly cloned"
                ci_var = cloned.variables[i_var.name]
                cloned.create_link_with_op(
                    variable=ci_var, op0=ci_var.source_op, op1=c_op
                )
            for o_var in op.outputs:
                assert (
                    o_var.name in cloned.variables
                ), f"Graph Copy Error, Variable {o_var.name} is not correctly cloned"
                co_var = cloned.variables[o_var.name]
                c_op.outputs.append(co_var)
                co_var.source_op = c_op
            if isinstance(op, QuantableOperation):
                for cfg, var in op.config_with_variable:
                    config_dict[cfg._hash] = (op, var)

        # relink config to there cloned master.
        for c_op in cloned.operations.values():
            if isinstance(c_op, QuantableOperation):
                for cfg, var in c_op.config_with_variable:
                    if cfg.dominated_by != cfg:
                        assert (
                            cfg.dominated_by._hash in config_dict
                        ), "Graph Copy Error, can not find a corresponding master config."
                        op, var = config_dict[cfg.dominated_by._hash]

                        op = cloned.operations[op.name]
                        assert isinstance(
                            op, QuantableOperation
                        ), "Graph Copy Error, Unexpected Master Operation Type."
                        for mcfg, mvar in op.config_with_variable:
                            if mvar.name == var.name:
                                cfg._dominator = mcfg

        # recreate input, output
        for name in self.inputs:
            cloned.inputs[name] = cloned.variables[name]
        for name in self.outputs:
            cloned.outputs[name] = cloned.variables[name]

        # check integrity
        for op in self.operations.values():
            if op.name not in cloned.operations:
                raise KeyError(f"Graph Copy Error, Operation {op.name} is Missing")
        for var in self.variables.values():
            if var.name not in cloned.variables:
                raise KeyError(f"Graph Copy Error, Variable {var.name} is Missing")
        for name in self.inputs:
            if name not in cloned.inputs:
                raise KeyError(f"Graph Copy Error, Input {name} is Missing")
        for name in self.outputs:
            if name not in cloned.outputs:
                raise KeyError(f"Graph Copy Error, Output {name} is Missing")
        cloned._num_of_generated_op = self._num_of_generated_op
        cloned._num_of_generated_var = self._num_of_generated_var
        cloned._detail = self._detail.copy()
        return cloned


class GraphBuilder(metaclass=SingletonMeta):
    r"""A base singleton for parser."""

    @abstractmethod
    def build(self, model_object: Any, **kwargs) -> BaseGraph:
        """Build BaseGraph from model_object, this function should be implemented"""


class GraphExporter(metaclass=SingletonMeta):
    r"""A base singleton for exporter."""

    @staticmethod
    def export_quantization_config(graph: BaseGraph):
        """Export Tensor Quantization Config (TQC) to file (JSON)."""

        render_buffer = {"configs": {}, "dispatchings": {}, "values": {}}
        # Render quantization config.
        for operation in graph.operations.values():
            op_dict = {}
            if isinstance(operation, QuantableOperation):
                for config, var in operation.config_with_variable:
                    op_dict[var.name] = dict(
                        bit_width=config.num_of_bits,
                        policy=config.policy.to_dict(),
                        state=config.state.name,
                        quant_min=config.quant_min,
                        quant_max=config.quant_max,
                        hash=hash(config),
                        dominator=hash(config.dominated_by),
                    )
                    if config.dominated_by == config:
                        if config.state != QuantizationStates.FP32:
                            render_buffer["values"][hash(config)] = {
                                "scale": _convert(config.scale),
                                "zero_point": _convert(config.offset),
                            }
            render_buffer["configs"][operation.name] = op_dict
            render_buffer["dispatchings"][operation.name] = operation.precision.name
        return render_buffer

    def dump_quantization_config(
        self, config_path: str | os.PathLike, graph: BaseGraph
    ):
        """Export Tensor Quantization Config (TQC) to file (JSON)."""
        render_buffer = self.export_quantization_config(graph)
        with open(file=config_path, mode="w", encoding="utf-8") as file:
            json.dump(render_buffer, file, indent=4)

    @abstractmethod
    def export(
        self,
        file_path: str | os.PathLike,
        graph: BaseGraph,
        config_path: Optional[str] = None,
        **kwargs,
    ):
        """Export BaseGraph to file_path, and export config to config_path if provided.
        This function should be implemented.
        """


class OperationExporter(metaclass=SingletonMeta):
    r"""A base singleton to transform an operation when exporting."""

    @abstractmethod
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        """Transform an operation to a new operation."""



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com
"""

import os
from typing import Optional, Union

import numpy as np
import onnx
import torch
from onnx import numpy_helper
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_operatorsetid,
    make_tensor,
    make_tensor_value_info,
)

import mppq
from mppq.common import GRAPH_OPSET_ATTRIB, ONNX_EXPORT_OPSET, ONNX_VERSION
from mppq.data import DataType, convert_any_to_numpy
from mppq.ir.base.graph import BaseGraph, GraphExporter, OperationExporter
from mppq.ir.base.opdef import Operation, Variable


class ConstantOfShapeExporter(OperationExporter):
    """
    PATCH 20211203, ConstantOfShape Op causes an export error.
    这一问题是由 ConstantOfShape 中的 value 格式问题引发的，下面的代码将导出正确的格式
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        op.attributes["value"] = numpy_helper.from_array(op.attributes["value"])
        return op


class MMCVExporter(OperationExporter):
    """mmcv operation must have a domain attribute."""

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        op.attributes["domain"] = "mmcv"
        return op


class InterpExporter(OperationExporter):
    """
    PATCH 20211216, interp op can not export input_shape attribute.
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        op.attributes.pop("input_shape")
        return op


class ORTExporter(OperationExporter):
    """onnxruntime operation must have a domain attribute."""

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        op.attributes["domain"] = "com.microsoft"
        return op


OP_CONVERTERS = {
    "ConstantOfShape": ConstantOfShapeExporter,
    "MMCVRoiAlign": MMCVExporter,
    "grid_sampler": MMCVExporter,
    "Interp": InterpExporter,
    "Attention": ORTExporter,
    "QAttention": ORTExporter,
    "QGemm": ORTExporter,
    "QLinearAdd": ORTExporter,
    "QLinearAveragePool": ORTExporter,
    "QLinearConcat": ORTExporter,
    "QLinearConv": ORTExporter,
    "QLinearGlobalAveragePool": ORTExporter,
    "QLinearLeakyRelu": ORTExporter,
    "QLinearMul": ORTExporter,
    "QLinearReduceMean": ORTExporter,
    "QLinearSigmoid": ORTExporter,
}


def _convert_value(value: Union[int, float, np.ndarray, torch.Tensor]):
    if isinstance(value, (int, float)):
        return value
    else:
        x = convert_any_to_numpy(value, accept_none=True)
        if x is None:
            return
        return x.tolist()


class OnnxExporter(GraphExporter):
    """
    PPQ 可以将 计算图 导出成 Onnx 标准格式，Onnx Exporter 不会导出 QDQ 节点。
    如需导出带有 QDQ 节点的 Onnx 文件，用户需要使用 OnnxRuntime Exporter

    任何导出器的导出逻辑都是原地进行的，它们将对传入的计算图对象进行原地修改，因此在导出之前你需要手动克隆计算图。
    """

    def export_graph(self, graph: BaseGraph) -> onnx.ModelProto:
        """
        Convert a PPQ IR to Onnx IR.
        This export will only convert PPQ Op and var to onnx, all quantization configs
        will be skipped.

        This function will try to keep the opset version of your graph unchanged.
        However if the opset is not given, ppq will convert it to with the global
        parameter ppq.core.ONNX_EXPORT_OPSET.
        """

        name = graph.name or f"mppq - v({mppq.__version__})"
        # Ready to export onnx graph definition.
        _inputs, _outputs, _initilizers, _nodes, _value_info = [], [], [], [], []

        # before we can export them, we firstly convert all ops to proper format.
        for op in graph.topological_sort():
            if op.type in OP_CONVERTERS:
                exporter = OP_CONVERTERS[op.type]()
                assert isinstance(
                    exporter, OperationExporter
                ), f"Expected an OpExporter here, however {type(exporter)} was given."
                op = exporter.export(op=op, graph=graph)

        for op in graph.topological_sort():
            _nodes.append(self.build_operator_proto(op))

        for variable in graph.variables.values():
            if variable.name in graph.inputs:
                var_shape = variable.shape
                onnx_dtype = variable.dtype.value
                _inputs.append(
                    make_tensor_value_info(
                        name=variable.name, elem_type=onnx_dtype, shape=var_shape
                    )
                )
            if variable.name in graph.outputs:
                var_shape = variable.shape
                onnx_dtype = variable.dtype.value
                _outputs.append(
                    make_tensor_value_info(
                        name=variable.name, elem_type=onnx_dtype, shape=var_shape
                    )
                )
            else:
                tensor_proto = self.build_variable_proto(variable)
                if not tensor_proto:
                    continue
                if variable.is_parameter:
                    _initilizers.append(tensor_proto)
                else:
                    _value_info.append(tensor_proto)

        graph_def = make_graph(
            name=name,
            nodes=_nodes,
            inputs=_inputs,
            outputs=_outputs,
            initializer=_initilizers,
            value_info=_value_info,
        )

        # if opset is missing from your graph, give it a default one.
        if GRAPH_OPSET_ATTRIB not in graph.extension_attrib:
            opsets = [make_operatorsetid("", ONNX_EXPORT_OPSET)]
        else:
            opsets = []
            for opset in graph.extension_attrib[GRAPH_OPSET_ATTRIB]:
                opsets.append(make_operatorsetid(opset["domain"], opset["version"]))
        onnx_model = make_model(
            graph_def,
            producer_name="mppq",
            ir_version=graph.extension_attrib.get("ir_version", ONNX_VERSION),
            opset_imports=opsets,
        )
        return onnx_model

    def build_operator_proto(self, operation: Operation) -> onnx.NodeProto:
        """
        Convert PPQ Op to Onnx Operation
        An Op consumes zero or more Tensors, and produces zero or more Tensors.
        """
        attributes = operation.attributes
        for key in attributes:
            value = attributes[key]
            if isinstance(value, DataType):
                attributes[key] = value.value
            if isinstance(value, torch.Tensor):
                if value.numel() == 0:
                    attributes[key] = None
                elif value.numel() == 1:
                    # convert to 1d array
                    attributes[key] = convert_any_to_numpy([value.item()])
                else:
                    attributes[key] = convert_any_to_numpy(value)

        op_proto = make_node(
            op_type=operation.type,
            inputs=[_.name for _ in operation.inputs],
            outputs=[_.name for _ in operation.outputs],
            name=operation.name,
            **attributes,
        )

        return op_proto

    def build_variable_proto(self, variable: Variable) -> onnx.TensorProto | None:
        """
        Convert PPQ Variable to Onnx TensorProto, There are 2 different types of Tensor
        in Onnx:
            Variable: Represents a Tensor whose value is not known until inference-time.
            Constant: Represents a Tensor whose value is known.
        """
        # Parameter Variable in PPQ, Constant Variable in Onnx
        if not variable.has_value:
            return

        var_shape = variable.value.shape
        pytorch_dtype = variable.value.dtype
        onnx_dtype = DataType.from_torch(pytorch_dtype).value
        value = variable.value
        is_raw_format = False
        if value.numel() == 0:
            value = []
        elif value.ndim == 0:  # Pytorch Scalar Type
            value = [value.item()]
        elif value.ndim >= 1:
            value = convert_any_to_numpy(value, False).flatten()
            value = value.tobytes()
            is_raw_format = True
        tensor_proto = make_tensor(
            name=variable.name,
            data_type=onnx_dtype,
            dims=var_shape,
            vals=value,
            raw=is_raw_format,
        )
        return tensor_proto

    def export(
        self,
        file_path: str | os.PathLike,
        graph: BaseGraph,
        config_path: Optional[str] = None,
        save_as_external_data: bool = False,
        **kwargs,
    ):
        # if a valid config path is given, export quantization config to there.
        if config_path is not None:
            self.dump_quantization_config(config_path, graph)

        size_threshold = 0 if save_as_external_data else 1024
        model_pb = self.export_graph(graph=graph)
        onnx.save_model(
            model_pb,
            file_path,
            save_as_external_data=save_as_external_data,
            size_threshold=size_threshold,
            convert_attribute=True,
        )



from copy import deepcopy
from typing import Any, List, Optional, Protocol, Tuple

from torch import Tensor

from mppq.data import convert_any_to_tensor
from mppq.ir.base.opdef import Operation, Variable
from mppq.quant import (
    OperationQuantizationConfig,
    QuantizationStates,
    TargetPrecision,
    TensorQuantizationConfig,
)


class BaseQuantFunction(Protocol):
    def __call__(self, tensor: Tensor, config: TensorQuantizationConfig) -> Tensor:
        return tensor


class QuantableOperation(Operation):
    r"""Quantable Operation (量化算子) 用来表示一个已经被量化了的算子
    相比于普通算子，一个量化算子具有以下额外的功能

    1. 每一个量化算子都将具有一个 config(OperationQuantizationConfig) 属性
       PPQ 使用这个东西描述量化细节，在整个网络中，有且只有这一个量化表示
       executor, dispatcher, optimization pass, exporter都是围绕这一属性工作的

    2. 每一个量化算子都将有一个 dequantize 方法和 restore_quantize_state 方法，
       一旦一个量化算子被 dequantize() 方法解除量化，该算子的 OperationQuantizationConfig 将被修改状态
       从而使得该算子的输入输出量化被暂时停用
       被解除量化的算子可以随时通过 restore_quantize_state 方法恢复量化状态
       对一个算子多次重复执行 dequantize 是可以的

    3. 每一个量化算子都将有一个 baking parameter 方法
       当算子具有有效的量化参数时，baking_parameters() 方法将对该算子的参数执行静态量化
       一旦静态量化完成，算子参数将被量化后的值替换；同时 config 的状态将被设置为: baked

    4. 每一个量化算子都将有一个 store_parameter_value 方法
       该方法将算子目前的参数保存入缓存；PPQ 将在创建 QuantableOperation 时执行此函数
       从而保存算子的原始参数，以备后续取用。
       一个显而易见的例子是，一旦算子执行了 baking_parameters 方法，它的参数值将被修改，
       此时若要完全还原算子状态，需要从缓存中取出算子的原始参数，并替换当前的值
       当你调用 restore_quantize_state 时，该方法会从缓存中取回保存的参数值并执行替换
       你不应当手动调用该方法，该方法将影响到 PPQ 的核心逻辑正确性

    5. 一个量化算子是可拷贝的，该拷贝只会拷贝算子的基本信息以及绑定的 OperationQuantizationConfig
    """

    def __init__(
        self,
        convert_from: Operation,
        quantize_config: OperationQuantizationConfig,
        platform: TargetPrecision,
    ):
        # Simply copy all attributes from fp32 operation
        # inputs, outputs will be created by QuantableGraph
        super().__init__(
            op_type=convert_from.type,
            inputs=convert_from.inputs.copy(),
            outputs=convert_from.outputs.copy(),
            attributes=convert_from.attributes,
            name=convert_from.name,
            precision=platform,
            opset=convert_from.opset,
        )
        self.config = quantize_config
        self._dequantized = False

    @property
    def input_quant_config(self) -> List[TensorQuantizationConfig]:
        return self.config.input_quantization_config

    @property
    def output_quant_config(self) -> List[TensorQuantizationConfig]:
        return self.config.output_quantization_config

    def baking_parameters(self, quant_func: BaseQuantFunction):
        for config, var in self.config_with_variable:
            if var.is_parameter and config.state in (
                QuantizationStates.ACTIVATED,
                QuantizationStates.PASSIVE,
            ):
                assert isinstance(var, QuantableVariable)
                assert len(var.dest_ops) == 1, (
                    f", Parameter {var.name} has {len(var.dest_ops)} destinations, "
                    "Baking parameter that has more than 1 destinations will incur "
                    "unexpected problems, PPQ does not support parameters with more "
                    "than 1 related operation, reform your graph first."
                )
                var.value = quant_func(var.value, config)
                if config.state == QuantizationStates.ACTIVATED:
                    config.state = QuantizationStates.BAKED
                if config.state == QuantizationStates.PASSIVE:
                    config.state = QuantizationStates.PASSIVE_BAKED
        return self

    def store_parameter_value(self):
        for var in self.inputs + self.outputs:
            if var.is_parameter:
                assert isinstance(var, QuantableVariable)
                # convert var.value to torch.Tensor
                # notice here we set device = None, this conversion will not change
                # var.value.device anyway. So that we can use var.value.device as a
                # deploy device for stored_value
                var.stored_value = convert_any_to_tensor(
                    var.value, device="cpu"
                ).clone()
        return self

    def dequantize(
        self, parameter_only: bool = False, expire_device: Optional[str] = "cpu"
    ):
        if self._dequantized:
            return self
        for var, quant_config in zip(
            self.inputs + self.outputs,
            self.config.input_quantization_config
            + self.config.output_quantization_config,
        ):
            if parameter_only and not var.is_parameter:
                continue
            quant_config.detail["Stored State"] = quant_config.state
            assert isinstance(var, QuantableVariable)
            if var.is_parameter:
                # convert var.value to torch.Tensor
                # notice here we set device = None, this conversion will not change
                # var.value.device anyway. So that we can use var.value.device as a
                # deploy device for stored_value
                stored_value = convert_any_to_tensor(var.value, device=expire_device)
                var.value = convert_any_to_tensor(var.value, device=None)
                var.value = convert_any_to_tensor(
                    var.stored_value,
                    device=var.value.device if var.has_value else None,
                )
                var.stored_value = stored_value
            quant_config.state = QuantizationStates.FP32
        self._dequantized = True
        return self

    def restore_quantize_state(self, expire_device: Optional[str] = "cpu"):
        if not self._dequantized:
            return self
        for var, quant_config in zip(
            self.inputs + self.outputs,
            self.config.input_quantization_config
            + self.config.output_quantization_config,
        ):
            if "Stored State" in quant_config.detail:
                quant_config.state = quant_config.detail["Stored State"]
                quant_config.detail.pop("Stored State")
                if var.is_parameter:
                    assert isinstance(var, QuantableVariable)
                    stored_value = convert_any_to_tensor(
                        var.value, device=expire_device
                    )
                    var.value = convert_any_to_tensor(var.value, device=None)
                    var.value = convert_any_to_tensor(
                        var.stored_value,
                        device=var.value.device if var.has_value else None,
                    )
                    var.stored_value = stored_value
        self._dequantized = False
        return self

    @property
    def config_with_variable(self) -> List[Tuple[TensorQuantizationConfig, Variable]]:
        """Just a helper function, This function will list all related config
        and variable with current operation.

        Returns:
            List[Tuple[TensorQuantizationConfig, Variable]]: [description]
        """
        ret = []
        for cfg, var in zip(self.config.input_quantization_config, self.inputs):
            ret.append((cfg, var))
        for cfg, var in zip(self.config.output_quantization_config, self.outputs):
            ret.append((cfg, var))
        return ret

    def __deepcopy__(self, memo=None) -> "QuantableOperation":
        return QuantableOperation(
            convert_from=deepcopy(self, memo),
            quantize_config=deepcopy(self.config, memo),
            platform=self.precision,
        )


class QuantableVariable(Variable):
    def __init__(self, convert_from: Variable) -> None:
        super().__init__(
            name=convert_from.name,
            dest_ops=convert_from.dest_ops.copy(),
            source_op=convert_from._source_op,
            value=convert_from._value,
            is_parameter=convert_from.is_parameter,
            shape=convert_from._shape,
            dtype=convert_from._dtype,
        )
        self._fp32_value = None
        if convert_from.has_value:
            self._fp32_value = convert_any_to_tensor(convert_from.value, device="cpu")

    @property
    def stored_value(self) -> Any:
        return self._fp32_value

    @stored_value.setter
    def stored_value(self, value: Any):
        self._fp32_value = value

    @property
    def dest_op_configs(self) -> List[TensorQuantizationConfig]:
        _dest_op_configs, _dest_idx = [], self.dest_idx
        for idx, op in enumerate(self.dest_ops):
            if isinstance(op, QuantableOperation):
                _dest_op_configs.append(
                    op.config.input_quantization_config[_dest_idx[idx]]
                )
            else:
                _dest_op_configs.append(None)
        return _dest_op_configs

    @property
    def source_op_config(self) -> None | TensorQuantizationConfig:
        if isinstance(self.source_op, QuantableOperation) and self.src_idx is not None:
            return self.source_op.config.output_quantization_config[self.src_idx]
        return None

    def copy(self, copy_value: bool = False):
        clone = QuantableVariable(super().copy(copy_value))
        if copy_value and self._fp32_value is not None:
            clone._fp32_value = self._fp32_value.clone()
        else:
            clone._fp32_value = self._fp32_value
        return clone



"""
This File Contains Legacy Dispatchers.
Refer to ppq.scheduler.perseus for updated implementation.
"""

from typing import Collection, Dict

from mppq.dispatcher.base import (
    DISPATCHER_TABLE,
    GraphDispatcher,
    reverse_tracing_pattern,
    soi_generators,
    soi_receivers,
    value_tracing_pattern,
)
from mppq.ir.base.graph import BaseGraph
from mppq.ir.search import SearchableGraph
from mppq.quant import TargetPrecision


@DISPATCHER_TABLE.register("conservative")
class ConservativeDispatcher(GraphDispatcher):
    r"""Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable
    operations, Shape-Or-Index related operations and remaining operations, all sub
    classes of GraphDispatcher will give an implementation of function "dispatch" to
    send all operations to their proper platform.

    Conservative Dispatcher cuts graph in a conservative way, which means it takes as
    much as possible operations into fp32 platform.

    ATTENTION:

        platform attribute will greatly affect quantizer's quantization logic, and the
        execution result.

        - If operation is sent to a quantable platform, then its inputs and outputs
          will be quantized if necessary.
        - If operation is classified as shape-or-index related operation, then its
          execution will be taken with cpu.
        - If operation is sent to a fp32 platform, then its inputs and outputs shall
          never be quantized.

    ATTENTION:

        this dispatcher will insert necessary DeviceSwitch operations between
        shape-or-index operations and others.
    """

    def __init__(self, graph: BaseGraph) -> None:
        self.graph = graph

    def dispatch(
        self,
        quant_types: Collection[str],
        quant_precision: TargetPrecision,
        fp32_platform: TargetPrecision = TargetPrecision.FP32,
        soi_platform: TargetPrecision = TargetPrecision.SOI,
        **kwargs,
    ) -> Dict[str, TargetPrecision]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.

        There are 3 default platform during dispatching:
            quant_precision - all quantable parts of graph will be dispatched to this
                             platform
            SOI_platform   - Aka. Shape or Index related operations will be dispatched
                             to this platform.
            fp32_platform  - there are some operations receiving results from both
                             quant_precision and SOI_platform, they will be dispatched
                             they will be dispatched to fp32_platform.

        ATTENTION:

            Quantization follows this dispatching, and only the operations within
            quantable platform will be quantized in the future.

        ATTENTION:

            this dispatcher will insert necessary DeviceSwitch operations between
            shape-or-index operations and others.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this
                dispatcher.
            quant_types(Set[str]): all quantable types for given platforms.
            quant_precision (TargetPlatform):
                platform object where quantable parts will goes to.
            fp32_platform (TargetPlatform):
                platform object where SOI parts will goes to.
            SOI_platform (TargetPlatform):
                platform object where remaining parts will goes to.
        """
        graph = self.graph
        receivers, generators = soi_receivers(graph), soi_generators(graph)
        search_engine, soi_operations = SearchableGraph(graph), set(receivers)

        quant_operations = search_engine.opset_matching(
            sp_expr=lambda op: op.is_computing_op,
            rp_expr=value_tracing_pattern,
            ep_expr=lambda op: (op.type not in quant_types) or op.is_boundary,
            direction="down",
        )
        quant_operations.filter(lambda op: op.type not in quant_types)

        computing_extensions = search_engine.opset_matching(
            sp_expr=lambda op: op.is_computing_op,
            rp_expr=value_tracing_pattern,
            ep_expr=lambda op: (
                op.type in {"Shape", "TopK", "NonMaxSuppression"} or op.is_boundary
            ),
            direction="down",
        )

        # we assume all 'Shape', 'NonMaxSuppression', 'ConstantOfShape', 'Topk'
        # operations are SOI generators.
        shape_forward_matching = search_engine.opset_matching(
            sp_expr=lambda op: op in generators and op.type not in {"Constant"},
            rp_expr=value_tracing_pattern,
            ep_expr=lambda op: (
                op in receivers
                or op in quant_operations
                or op.is_boundary
                or op.is_computing_op
            ),
            direction="down",
        )

        # remove computing operations and quant operations from matching
        shape_forward_matching.filter(
            lambda op: op.is_computing_op or op in quant_operations
        )

        # update matchings, ready for further searching.
        soi_operations.update(shape_forward_matching)

        while True:
            # there are some particular cases where a single matching can not handle.
            # to cover all shape-related operations, a reverse matching is required.
            shape_backward_matching = search_engine.opset_matching(
                sp_expr=lambda op: op in soi_operations and op.type != "Shape",
                rp_expr=reverse_tracing_pattern,
                ep_expr=lambda op: (
                    op in soi_operations
                    or op in quant_operations
                    or op.is_boundary
                    or op.is_computing_op
                ),
                direction="up",
            )

            # remove computing operations and quant operations from matching
            shape_backward_matching.filter(
                lambda op: op.is_computing_op or op in quant_operations
            )

            if all([(op in soi_operations) for op in shape_backward_matching]):
                break

            # update matchings
            soi_operations.update(shape_backward_matching)

        # generate dispatching table.
        dispatching_table = {}
        for operation in graph.operations.values():
            if operation in soi_operations and operation not in computing_extensions:
                dispatching_table[operation.name] = soi_platform
            elif operation in quant_operations:
                dispatching_table[operation.name] = quant_precision
            else:
                dispatching_table[operation.name] = fp32_platform

        for operation in graph.operations.values():
            # move Topk, Shape, NonMaxSuppression to the platform same as their input.
            if operation.type in {"Shape", "TopK", "NonMaxSuppression"}:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[
                        source_op.name
                    ]
                else:
                    dispatching_table[operation.name] = fp32_platform

            # move activations to the platform same as their input.
            if operation.is_linear_activation:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[
                        source_op.name
                    ]

        return dispatching_table



from mppq.executor.base import BaseGraphExecutor, QuantRuntimeHook, RuntimeHook
from mppq.executor.torch import TorchExecutor, TorchQuantizeDelegator

__all__ = [
    "BaseGraphExecutor",
    "QuantRuntimeHook",
    "RuntimeHook",
    "TorchExecutor",
    "TorchQuantizeDelegator",
]



"""
Copyright Wenyi Tang 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

Config pytest
"""

from pathlib import Path

import pytest

model_dir = Path(__file__).parent.parent / "models"


def pytest_generate_tests(metafunc: pytest.Metafunc):
    """Generate parametrized arguments to all tests with arg 'model'."""

    models = sorted(model_dir.rglob("*.onnx"))
    if "model" in metafunc.fixturenames:
        metafunc.parametrize("model", models)



from collections import defaultdict
from contextlib import suppress
from typing import Callable, Dict, Iterable, List, Literal, Optional, Union

import torch
from tqdm import tqdm

from mppq.data import convert_any_to_numpy
from mppq.executor import TorchExecutor
from mppq.executor.base import BaseGraphExecutor
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.quantize import QuantableOperation
from mppq.ir.deploy import QuantableGraph
from mppq.quantization.analyse.util import MeasurePrinter, MeasureRecorder
from mppq.utils.fetch import tensor_random_fetch


def layerwise_error_analyse(  # noqa: C901
    executor: BaseGraphExecutor,
    dataloader: Iterable,
    interested_outputs: Optional[Union[str, List[str]]] = None,
    collate_fn: Optional[Callable] = None,
    method: Literal["cosine", "snr", "mse"] = "snr",
    reduce: Literal["mean", "sum"] = "mean",
    steps: int = 8,
    verbose: bool = True,
) -> Dict[str, float]:
    """Measure the quantization error of each operation A dictionary contains
    output differences for all operation will be returned as a result.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}

    if verbose is set as True, this function will display error report at last.

    The key of the dictionary is an operation name while the value of corresponding key
        is the difference between quantized output and float output of this operation.

    Result {'operation name 1': 0.933} means quantizing operation 1
        will generates 0.933 quantization error to output variable

    ATTENTION: Output difference is measured at operation-level.

    Args:
        graph (BaseGraph):
            A fully quantized graph instance.

        dataloader (Iterator):
            Test dataloader, this function will measure quantization error based on
            given data.

        collate_fn (Callable, optional):
            An data preprocessing function provided by user to convert data from
            dataloader towards executable format. If set as None, then no action will
            be taken during preprocessing.

        method (str, optional):
            A string indicates a measurement to calculate the difference of quantized
            output and fp32 one. 'cosine', 'snr', and 'mse' is supported in PPQ for now.

        steps (Int, optional)
            computation steps.

        interested_outputs (Union[str, List[str]] = None)
            a list contains your interested output variables. if set as None, all graph
            output variables will be measured via this function.

    Returns:
        A dictionary contains output differences for all operation will be returned
        from this function.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}
    """

    graph = executor._graph  # pylint: disable=protected-access
    assert graph is not None
    if interested_outputs is None:
        interested_outputs = [name for name in graph.outputs]

    if isinstance(interested_outputs, str):
        interested_outputs = [interested_outputs]

    # find all quantable operations.
    quantable_operations = []
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.dequantize()

            # we only need reports from computing op.
            if operation.is_computing_op:
                quantable_operations.append(operation)

    # dequantize all operations, create recorder for each operation
    recorders = {}
    for operation in quantable_operations:
        if isinstance(operation, QuantableOperation):
            recorders[operation.name] = MeasureRecorder(
                measurement=method, reduce=reduce
            )

    # run for each quantable operations:
    for operation in tqdm(
        quantable_operations, desc="Analysing Layerwise quantization error:"
    ):
        assert isinstance(operation, QuantableOperation)
        recorder = recorders[operation.name]
        assert isinstance(recorder, MeasureRecorder)

        for idx, batch in enumerate(dataloader):
            if collate_fn is not None:
                batch = collate_fn(batch)
            fp_outputs = executor.forward(inputs=batch, output_names=interested_outputs)

            # manually override quantization state
            operation.restore_quantize_state()
            qt_outputs = executor.forward(inputs=batch, output_names=interested_outputs)

            for fp_output, qt_output in zip(fp_outputs, qt_outputs):
                recorder.update(y_pred=qt_output, y_real=fp_output)

            # manually override quantization state
            operation.dequantize()
            if idx >= steps:
                break

    # restore quantization states
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.restore_quantize_state()

    results: Dict[str, float] = {}
    for operation in quantable_operations:
        assert isinstance(operation, QuantableOperation)
        results[operation.name] = recorders[operation.name].measure

    if verbose:
        method_str = "MEASUREMENT"
        if method == "snr":
            method_str = "NOISE:SIGNAL POWER RATIO"
        if method == "cosine":
            method_str = "COSINE SIMILARITY"
        if method == "mse":
            method_str = f"{reduce.upper()} SQUARE ERROR"
        MeasurePrinter(
            results,
            order="large_to_small",
            measure=method_str,
            percentage=method in {"snr", "cosine"},
        ).print()
    return results


def variable_analyse(
    graph: BaseGraph,
    dataloader: Iterable,
    interested_outputs: Union[str, List[str]],
    collate_fn: Optional[Callable] = None,
    running_device="cuda",
    samples_per_step: int = 65536,
    steps: int = 8,
    dequantize: bool = False,
):
    if isinstance(interested_outputs, str):
        interested_outputs = [interested_outputs]

    quant_graph = QuantableGraph(graph)

    executor = TorchExecutor(graph=graph, device=running_device)
    if dequantize:
        quant_graph.dequantize_graph()

    data_collector = defaultdict(list)
    for idx, batch in enumerate(dataloader):
        if collate_fn is not None:
            batch = collate_fn(batch)
        fp_outputs = executor.forward(inputs=batch, output_names=interested_outputs)
        for output, output_name in zip(fp_outputs, interested_outputs):
            data_collector[output_name].append(
                tensor_random_fetch(
                    tensor=output, num_of_fetches=samples_per_step
                ).unsqueeze(0)
            )
        if idx >= steps:
            break

    for name in interested_outputs:
        # pylint: disable=import-outside-toplevel
        from matplotlib import pyplot as plt

        tensor = torch.cat(data_collector[name]).flatten()
        tensor = convert_any_to_numpy(tensor, False)

        plt.figure(figsize=(12, 8))
        plt.title(f"Histogram Result of Variable {name}:")
        plt.hist(tensor, bins=64)
        plt.show()

    if dequantize:
        quant_graph.restore_quantize_state()


def parameter_analyse(graph: BaseGraph):
    ranges, stds, means = {}, {}, {}
    for operation in graph.operations.values():
        for var in operation.parameters:
            value = var.value
            assert isinstance(value, torch.Tensor)
            if value.numel() <= 1:
                continue

            _min, _max, _std, _mean = 0, 0, 0, 0
            with suppress(Exception):
                _min = value.min().item()
                _max = value.max().item()
                _std = value.std().item()
                _mean = value.mean().item()

            ranges[f"{var.name}[{operation.name}]"] = _max - _min
            stds[f"{var.name}[{operation.name}]"] = _std
            means[f"{var.name}[{operation.name}]"] = abs(_mean)

    MeasurePrinter(ranges, order="large_to_small", measure="Value Range").print()
    MeasurePrinter(stds, order="large_to_small", measure="Value Std").print()
    MeasurePrinter(means, order="large_to_small", measure="Value Mean(Abs)").print()



# pylint: disable=invalid-name

from typing import Any, List, Optional, Protocol, Sequence, Tuple

import torch

from mppq.ir.base.opdef import Operation
from mppq.ir.base.quantize import QuantableOperation
from mppq.quant import TargetPrecision
from mppq.register import Registry


class TorchBackendContext:
    def __init__(self, executing_device: str) -> None:
        self.executing_device = executing_device


def ASSERT_NUM_OF_INPUT(
    op: Operation,
    values: Sequence[torch.Tensor],
    min_num_of_input: int = -1,
    max_num_of_input: int = 99,
):
    if min_num_of_input == max_num_of_input:
        if len(values) != min_num_of_input:
            raise ValueError(
                f"Can not feed value to operation {op.name}, "
                f"expects exact {min_num_of_input} inputs, "
                f"however {len(values)} was given"
            )
    elif len(values) > max_num_of_input:
        raise ValueError(
            f"Too many input value for {op.name}, "
            f"expects {max_num_of_input} inputs at most, "
            f"however {len(values)} was given"
        )
    elif len(values) < min_num_of_input:
        raise ValueError(
            f"Too few input value for {op.name}, "
            f"expects {min_num_of_input} inputs at least, "
            f"however {len(values)} was given"
        )


def GET_ATTRIBUTE_FROM_OPERATION(
    op: Operation, attribute: str, compulsive: bool = False, default: Any = None
):
    """Try to get an attribute from operation. If an attribute is compulsive,
    then operation must give a value of it, otherwise an error will be thrown.
    If an attribute is not compulsive, a default value will be given if
    operation.attributes do not holds a value of requesting attribute.

    Args:
        op (Operation): Operation instance.
        attribute (str): Attribute name.
        compulsive (bool): Whether is a compulsive attribute.
        default (Any, optional): [description]. default value of attribute.
    """
    if attribute in op.attributes:
        return op.attributes[attribute]
    else:
        if compulsive:
            raise KeyError(
                f"Operation {op.name} is supposed to have a value of attribute "
                f"{attribute}. However this value is missing from currecnt operation.",
            )
        else:
            return default


def GET_VALUE_FROM_INPUTS(
    values: Sequence[torch.Tensor], idx: int
) -> Optional[torch.Tensor]:
    assert isinstance(idx, int) and idx >= 0
    if len(values) > idx:
        return values[idx]
    else:
        return None


def ASSERT_IS_QUANT_OP(op):
    if not isinstance(op, QuantableOperation):
        raise TypeError(
            "Given Operation is expected as a QuantableOperation, "
            f"however {type(op)} was given."
        )


def FORCE_CONVERT_DEVICE(
    value: torch.Tensor, device: str | torch.device
) -> torch.Tensor:
    return value.to(device=device, copy=True)


def VALUE_TO_EXECUTING_DEVICE(
    op: Operation, ctx: Optional[TorchBackendContext], values: Sequence[torch.Tensor]
) -> List[torch.Tensor]:
    values = list(values)
    if ctx is None:
        device = values[0].device
    else:
        device = ctx.executing_device
    for idx, (plat, value) in enumerate(zip(op.socket.in_plat, values)):
        if value is None:
            continue
        if plat == TargetPrecision.SOI or op.precision == TargetPrecision.SOI:
            values[idx] = value.cpu()
        else:
            values[idx] = value.to(device)
    return values


class OperationForwardProtocol(Protocol):
    """A protocol for operation forward function."""

    def __call__(
        self,
        op: Operation,
        values: Sequence[torch.Tensor],
        ctx: Optional[TorchBackendContext] = None,
        **kwargs,
    ) -> torch.Tensor | Tuple[torch.Tensor, ...]:
        """
        Args:
            op: operation information (precision, attributes, types, etc.)
            values: input tensors
            ctx: execution device information

        Returns:
            a result tensor or a tuple of result tensors
        """
        raise NotImplementedError


DEFAULT_BACKEND_TABLE: Registry[OperationForwardProtocol] = Registry("BACKEND_TABLE")



from mppq.frontend.base import EXPORTER, PARSER
from mppq.frontend.onnx.onnx_exporter import OnnxExporter
from mppq.frontend.onnx.onnx_parser import OnnxParser
from mppq.frontend.onnx.onnxruntime_exporter import ONNXRUNTIMExporter
from mppq.frontend.onnx.openvino_exporter import OpenvinoExporter

__all__ = [
    "OnnxExporter",
    "OnnxParser",
    "ONNXRUNTIMExporter",
    "OpenvinoExporter",
]


PARSER.register("onnx")(OnnxParser)
EXPORTER.register("onnx")(OpenvinoExporter)



from abc import ABCMeta, abstractmethod
from typing import Collection, Dict, Set, Type

from mppq.common import INT_CALCULATION_OPS, SOI_DATA_GENERATOR_OPS
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.opdef import Operation
from mppq.quant import TargetPrecision
from mppq.register import Registry


class GraphDispatcher(metaclass=ABCMeta):
    r"""Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable
    operations, Shape-Or-Index (SOI) related operations and remaining operations, all
    sub classes of GraphDispatcher will give an implementation of function "dispatch"
    to send all operations to their proper platform.

    ATTENTION:

        platform attribute will greatly affect quantizer's quantization logic, and the
        execution result.
        If operation is sent to a quantable platform, then its inputs and outputs will
        be quantized if necessary.
        if operation is classified as shape-or-index related operation, then its
        execution will be taken with cpu.
        if operation is sent to a fp32 platform, then its inputs and outputs shall
        never be quantized.
    """

    @abstractmethod
    def __init__(self, graph: BaseGraph, **kwargs) -> None:
        raise NotImplementedError("Impl this first.")

    @abstractmethod
    def dispatch(
        self, quant_types: Collection[str], quant_precision: TargetPrecision, **kwargs
    ) -> Dict[str, TargetPrecision]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.
        """
        raise NotImplementedError("Impl this first.")

    def __call__(
        self, quant_types: Collection[str], **kwargs
    ) -> Dict[str, TargetPrecision]:
        return self.dispatch(quant_types, **kwargs)


def value_tracing_pattern(x: Operation, y: Operation) -> bool:
    if y.type in INT_CALCULATION_OPS:
        # shape can go through above operations as a input, under this circumstance,
        # their output should still be a tensor of shape.
        # However if shape was feed as a parameter for those operations, then their
        # outputs are irrelevant with shape flow.
        return y.inputs[0].source_op == x
    if y.type == "ScatterND":
        # ScatterND has 2 quant input.
        return y.inputs[0].source_op == x or y.inputs[-1].source_op == x

    if y.type in {"ConstantOfShape", "Shape", "NonMaxSuppression"}:
        # Inputs: (1)
        #   input : T
        #     1D tensor. The shape of the expected output tensor.
        #     If empty tensor is given, the output would be a scalar.
        #     All values must be >= 0.
        # see also:
        # https://github.com/onnx/onnx/blob/master/docs/Operators.md#ConstantOfShape
        return False

    return True


def reverse_tracing_pattern(x: Operation, y: Operation) -> bool:
    if y.type in ("Shape", "TopK"):
        return False
    if x.type in INT_CALCULATION_OPS:
        return y == x.inputs[0].source_op
    if x.type == "ScatterND":
        return y == x.inputs[0].source_op or y == x.inputs[-1].source_op
    if x.type in {"NonMaxSuppression", "Shape"}:
        # remove constant of shape from here can speed up.
        return False
    return True


def soi_receivers(graph: BaseGraph) -> Set[Operation]:
    r"""Get operations in the graph that one of inputs is SOI."""
    receivers = set()
    for operation in graph.operations.values():
        for idx, plat in enumerate(operation.socket.in_plat):
            if plat in (TargetPrecision.SOI, TargetPrecision.FP32):
                receivers.add(operation.inputs[idx].source_op)
    receivers.discard(None)
    return receivers


def soi_generators(graph: BaseGraph) -> Set[Operation]:
    r"""Get operations in the graph that one of outputs may be SOI."""
    return {i for i in graph.operations.values() if i.type in SOI_DATA_GENERATOR_OPS}


DISPATCHER_TABLE: Registry[Type[GraphDispatcher]] = Registry("DISPATCHER_TABLE")



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

Changelist:
- Fix: build variables with correct data type.
- Feat: set known shape on building variables.
"""

# pylint: disable=protected-access

import os
from contextlib import suppress
from itertools import chain
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Dict, Iterable, List

import numpy as np
import onnx
import onnx.version_converter as ovc
from onnx import helper, mapping, numpy_helper

from mppq.common import (
    DEFAULT_OPSET_DOMAIN,
    DEFAULT_OPSET_VERSION,
    GRAPH_OPSET_ATTRIB,
    ONNX_VERSION,
)
from mppq.data import DataType
from mppq.ir.base.graph import BaseGraph, GraphBuilder
from mppq.ir.base.opdef import Operation, Opset, Variable
from mppq.logger import nest


class OnnxParser(GraphBuilder):
    """Parse ONNX model to PPQ IR graph."""

    _log = nest("OnnxParser")

    def _build_variables(
        self,
        graph: BaseGraph,
        graph_inputs: List[str],
        graph_outputs: List[str],
        op_inputs: Dict[str, list],
        op_outputs: Dict[str, list],
        value_info: Dict[str, SimpleNamespace],
    ) -> BaseGraph:
        var_list = []

        for op_name, _ in graph.operations.items():
            for var_name in op_inputs[op_name]:
                var_list.append(var_name)
            for var_name in op_outputs[op_name]:
                var_list.append(var_name)

        # create all variable at once.
        for var_name in set(var_list):
            if var_name in value_info:
                var_dtype = DataType.from_numpy(value_info[var_name].dtype)
                var_shape = list(value_info[var_name].shape)
            else:
                var_dtype = DataType.FP32
                var_shape = None
            graph.variables[var_name] = Variable(
                name=var_name, dtype=var_dtype, shape=var_shape
            )

        # build graph's input, output variables.
        try:
            for var_name in graph_inputs:
                if var_name not in graph.variables:
                    continue
                graph.inputs[var_name] = graph.variables[var_name]
            for var_name in graph_outputs:
                graph.outputs[var_name] = graph.variables[var_name]
        except KeyError as e:
            raise KeyError(
                "seems you got an input/output variable that is not linked to any "
                "operation."
            ) from e

        # build operation inputs, outputs variables.
        for op in graph.operations.values():
            for var_name in op_inputs[op.name]:
                var = graph.variables[var_name]
                var.dest_ops.append(op)
                op.inputs.append(graph.variables[var_name])
            for var_name in op_outputs[op.name]:
                var = graph.variables[var_name]
                var.source_op = op
                op.outputs.append(graph.variables[var_name])
        return graph

    def _initialize_params(
        self, graph: BaseGraph, initializer: Dict[str, Any]
    ) -> BaseGraph:
        for var in graph.variables.values():
            if var.name in initializer:
                for dest_op in var.dest_ops:
                    assert isinstance(dest_op, Operation)
                    dest_op.parameters.append(var)
                var.value = initializer[var.name]
                var.is_parameter = True
        return graph

    def _de_inplace(self, graph: BaseGraph) -> BaseGraph:
        """Remove inplace layer in netdef if the names of bottom and top are same,
        it means the computation of this layer is in place."""

        def new_name(_name: str):
            if _name == "":
                return ""
            elif _name not in total_write_times or current_write_times:
                return _name
            elif current_write_times[_name] == total_write_times[_name]:
                return _name
            else:
                return f"{_name}_ver{current_write_times[_name]}"

        total_write_times: Dict[str, int] = {}
        for op in graph.operations.values():
            for top in op.outputs:
                total_write_times.setdefault(top.name, 0)
                total_write_times[top.name] += 1

        current_write_times = {}
        for name in graph.inputs.keys():
            total_write_times[name] = 0
            current_write_times[name] = 0

        for op in graph.operations.values():
            for bottom in op.inputs:
                if bottom.is_parameter:
                    continue
                bottom.name = new_name(bottom.name)
            for top in op.outputs:
                current_write_times.setdefault(top.name, 0)
                current_write_times[top.name] += 1
                top.name = new_name(top.name)
        return graph

    def _refine_graph(self, graph: BaseGraph) -> BaseGraph:
        for op in graph.operations.values():
            for key, value in op.attributes.items():
                if op.type == "Constant" or op.type == "ConstantOfShape":
                    # The attribute of 'Constant' node is a value, needs to convert to
                    # numpy array
                    assert isinstance(value, onnx.TensorProto)
                    value = numpy_helper.to_array(value).copy()
                if op.type == "Cast":
                    # The attribute of 'Cast' node is data type (represented in int),
                    # need to convert to numpy data type
                    assert isinstance(value, int)
                    value = helper.tensor_dtype_to_np_dtype(value)
                elif isinstance(value, bytes):
                    value = value.decode("utf-8")
                op.attributes[key] = value

        for input_var in list(graph.inputs.values()):
            # remove initializer from graph.inputs
            if input_var.has_value:
                graph.inputs.pop(input_var.name)
        return graph

    def _convert_opsets_to_str(self, opsets: Iterable) -> List[Dict[str, str]]:
        results = []
        for opset in opsets:
            results.append({"domain": opset.domain, "version": opset.version})
        return results

    def _build_graph(
        self,
        model_pb: onnx.ModelProto,
        opset: int,
        graph: BaseGraph,
        op_inputs_dict: Dict[str, List[str]],
        op_outputs_dict: Dict[str, List[str]],
    ):
        _rand_seed = 0  # used for name generation.
        for node in model_pb.graph.node:
            op_name = node.name
            if len(op_name) == 0:
                # some operation do not have a name, we just generate one.
                op_name = "generated_name_" + str(_rand_seed)
                _rand_seed += 1

            if op_name in graph.operations:
                raise KeyError(f"Duplicated operation {op_name} was found.")

            graph.operations[op_name] = Operation(
                name=op_name,
                op_type=node.op_type,
                attributes={
                    item.name: helper.get_attribute_value(item)
                    for item in node.attribute
                },
                opset=Opset(domain=DEFAULT_OPSET_DOMAIN, version=opset),
            )
            op_inputs_dict[op_name] = [var_name for var_name in node.input]
            op_outputs_dict[op_name] = [var_name for var_name in node.output]

    def build(
        self, model_object: str | os.PathLike | onnx.ModelProto, **kw
    ) -> BaseGraph:
        """Build PPQ IR graph from an onnx file."""
        if kw:
            self._log.warning("Extra argument is not accepted! Ignoring:")
            self._log.warning(f'\n  {"\n  ".join(kw.keys())}')
        if isinstance(model_object, onnx.ModelProto):
            model_pb = model_object
        elif not Path(model_object).exists():
            self._log.error(f"File {model_object} does not exist.")
            raise FileNotFoundError
        else:
            model_pb = onnx.load_model(model_object, load_external_data=False)
        if model_pb.ir_version < ONNX_VERSION:
            # try to upgrade onnx version to the latest
            with suppress(ovc.ConvertError):
                model_pb = ovc.convert_version(model_pb, DEFAULT_OPSET_VERSION)

        onnx.checker.check_model(model_pb)  # validate model
        opsets = model_pb.opset_import
        graph = BaseGraph(name=model_pb.graph.name)
        graph._detail[GRAPH_OPSET_ATTRIB] = self._convert_opsets_to_str(opsets)
        graph._detail["ir_version"] = model_pb.ir_version
        with suppress(onnx.shape_inference.InferenceError):
            model_pb = onnx.shape_inference.infer_shapes(model_pb, True)

        onnx_import_opset = DEFAULT_OPSET_VERSION
        for opset in graph._detail[GRAPH_OPSET_ATTRIB]:
            if opset["domain"] == DEFAULT_OPSET_DOMAIN or opset["domain"] == "":
                onnx_import_opset = opset["version"]
                break

        # a temporary storage for operation's inputs and outputs
        op_inputs_dict: Dict[str, List[str]] = {}
        op_outputs_dict: Dict[str, List[str]] = {}
        self._build_graph(
            model_pb, onnx_import_opset, graph, op_inputs_dict, op_outputs_dict
        )
        # value type info
        graph_pb = model_pb.graph
        value_info = {}
        for value in chain(graph_pb.input, graph_pb.output, graph_pb.value_info):
            if value.name in value_info:
                # this is not an error since input/output can be appeared in value info
                continue
            value_shape = [i.dim_value for i in value.type.tensor_type.shape.dim]
            value_dtype = mapping.TENSOR_TYPE_MAP[value.type.tensor_type.elem_type]
            value_info[value.name] = SimpleNamespace(
                shape=value_shape, dtype=value_dtype.np_dtype
            )

        initializer = {}
        for item in graph_pb.initializer:
            init_name = item.name
            value = numpy_helper.to_array(item)
            if value.dtype == np.float16:
                # we should cast half type back to float32 because most cpu kernels
                # don't support half type.
                value = value.astype("float32")
            initializer[init_name] = value

        inputs = [item.name for item in graph_pb.input]
        outputs = [item.name for item in graph_pb.output]
        graph = self._build_variables(
            graph,
            graph_inputs=inputs,
            graph_outputs=outputs,
            op_inputs=op_inputs_dict,
            op_outputs=op_outputs_dict,
            value_info=value_info,
        )
        graph = self._initialize_params(graph, initializer)
        graph = self._de_inplace(graph)
        graph = self._refine_graph(graph)
        return graph



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

PPQ API.

Users are always encouraged to use functions and object within this package.
推荐用户仅使用该包内的函数和对象。
"""

from mppq import __version__
from mppq.api.extension import register_operation, register_platform
from mppq.api.interface import (
    dispatch_graph,
    export_graph,
    export_onnx_graph,
    format_graph,
    load_graph,
    load_onnx_graph,
    load_quantizer,
    quantize,
)
from mppq.dispatcher.base import GraphDispatcher
from mppq.executor.torch import TorchExecutor
from mppq.ffi import ENABLE_CUDA_KERNEL
from mppq.frontend.base import GraphBuilder, GraphExporter
from mppq.logger import set_level as set_log_level
from mppq.quantizer.base import BaseQuantizer

__all__ = [
    "__version__",
    "TorchExecutor",
    "ENABLE_CUDA_KERNEL",
    "set_log_level",
    "register_operation",
    "register_platform",
    "dispatch_graph",
    "export_graph",
    "export_onnx_graph",
    "format_graph",
    "load_graph",
    "load_onnx_graph",
    "load_quantizer",
    "quantize",
    "GraphDispatcher",
    "GraphBuilder",
    "GraphExporter",
    "BaseQuantizer",
]

print(
    r"""
              ____  ____  ____
   ____ ___  / __ \/ __ \/ __ \
  / __ `__ \/ /_/ / /_/ / / / /
 / / / / / / ____/ ____/ /_/ /
/_/ /_/ /_/_/   /_/    \___\_\

"""
)



from itertools import cycle
from math import ceil
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

import torch
import torch.nn.functional as F
from tqdm import tqdm

from mppq.executor import BaseGraphExecutor
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.opdef import Operation, Variable
from mppq.ir.base.quantize import BaseQuantFunction, QuantableOperation
from mppq.ir.search import Path, SearchableGraph, TraversalCommand
from mppq.logger import debug
from mppq.quant import QuantizationProperty, QuantizationStates
from mppq.quantization.measure import torch_mean_square_error
from mppq.quantization.observer.base import (
    OBSERVER_TABLE,
    CalibrationHook,
    OperationObserver,
)
from mppq.quantization.optim.base import OPTIM_ALGORITHMS, QuantizationOptimizationPass
from mppq.utils.qfunction import ppq_fake_quant

OPTIMIZATION_LAYERTYPE_CONFIG = {
    "Relu",
    "MaxPool",
    "GlobalMaxPool",
    "PRelu",
    "AveragePool",
    "GlobalAveragePool",
}
EQUALIZATION_OPERATION_TYPE = {
    "Conv",
    "Gemm",
    "ConvTranspose",
}  # support all computing op types


@OPTIM_ALGORITHMS.register()
class SSDEqualizationPass(QuantizationOptimizationPass):
    r"""PPQ Custimized Layerwise Equalization Pass.

    This is another layerwise equalization pass which takes quantization error into
    consideration, for more details of equalization, please refer to
    LayerwiseEqualizationPass

    Equalization algo will only take effect when the estimated loss is less than
    loss_threshold * original_loss

    Compared with LayerwiseEqualizationPass, it guarantees a better result than the
    original model while consuming more time for loss estimation
    """

    def __init__(
        self,
        channel_ratio: float = 0.5,
        loss_threshold: float = 0.8,
        layer_norm: bool = False,
        quant_func: BaseQuantFunction = ppq_fake_quant,
        iteration: int = 3,
    ):
        r"""SSD Equalization Pass With Loss Checking.

        Args:
            channel_ratio (float, optional): all values below this ratio of maximum
                value of corresponding weight will be clipped to this ratio when
                calculating scale. Defaults to 0.5.

            loss_threshold (float, optional): the estimated loss should be below this
                threshold * original_loss for algo to take effect. Defaults to 0.8.

            layer_norm (bool, optional): whether to apply weight normalization.
                Defaults to False.

            quant_func (BaseQuantFunction, optional): quantization function.
                Defaults to ppq_fake_quant.

            iteration (int, optional): num of iterations to run, usually 3 would be
                enough. Defaults to 3.
        """
        super().__init__(name="SSD Equalization Pass")
        self.channel_ratio = channel_ratio
        self.loss_threshold = loss_threshold
        self.layer_norm = layer_norm
        self.quant_func = quant_func
        self.start_op_types = EQUALIZATION_OPERATION_TYPE
        self.relay_op_types = OPTIMIZATION_LAYERTYPE_CONFIG
        self.end_op_types = EQUALIZATION_OPERATION_TYPE
        self.iteration = iteration

    def collect_all_pairs(self, graph: BaseGraph) -> List[List[Operation]]:
        def limitation(path: Path) -> bool:
            for op in path.tolist()[:-1]:
                if len(graph.get_downstream_operations(op)) != 1:
                    return False
            return True

        search_engine = SearchableGraph(graph)
        forward_matchings = search_engine.process(
            TraversalCommand(
                sp_expr=lambda op: op.type in self.start_op_types,
                rp_expr=lambda x, y: y.type in self.relay_op_types,
                ep_expr=lambda op: op.type in self.end_op_types,
                direction="down",
            )
        )
        return [path.tolist() for path in forward_matchings if limitation(path)]

    def collect_activation_range(
        self,
        pair: List[Operation],
        executor: BaseGraphExecutor,
        data_loader: Iterable,
        collate_fn: Optional[Callable[[Any], Any]] = None,
        calib_steps: int = 1,
    ) -> Dict[Operation, torch.Tensor]:
        """Collect activation ranges for Conv ops in the pair.

        Args:
            pair (List[Operation]): equalization pair
            executor (BaseGraphExecutor): graph executor
            data_loader (Iterable): data loader
            collate_fn (Callable): batch func
            calib_steps (int): calibration steps

        Returns:
            Dict[Operation, torch.Tensor]: activation range of Conv ops in the pair
        """
        op_act_ranges = {
            op: torch.tensor(0) for op in pair if op.type in self.start_op_types
        }
        output_names = [op.outputs[0].name for op in op_act_ranges]
        for step, data in enumerate(cycle(data_loader)):
            if step >= calib_steps:
                break
            if collate_fn is not None:
                data = collate_fn(data)
            outputs = executor.forward(inputs=data, output_names=output_names)
            for op in op_act_ranges:
                op_conv_data = outputs[output_names.index(op.outputs[0].name)]
                op_conv_data_relu = F.relu(op_conv_data)  # take abs value
                op_conv_data_relu = op_conv_data_relu.permute(
                    1, 0, *(range(op_conv_data_relu.ndim)[2:])
                ).contiguous()
                op_conv_data_relu = op_conv_data_relu.reshape(
                    (op_conv_data_relu.shape[0], -1)
                )
                op_conv_data_relu = op_conv_data_relu.max(1)[0]
                op_act_ranges[op] = op_conv_data_relu + op_act_ranges[op]
        for op in op_act_ranges:
            op_act_ranges[op] = op_act_ranges[op] / calib_steps

        for op, op_range in op_act_ranges.items():
            assert isinstance(op_range, torch.Tensor)
            op_range_max = op_range.max()
            adjust_range = torch.where(
                op_range < op_range_max * self.channel_ratio,
                op_range_max * self.channel_ratio,
                op_range,
            )
            op_act_ranges[op] = adjust_range
        return op_act_ranges

    def layer_weight_norm(self, pairs: List[List[Operation]]):
        for pair in pairs:
            op_first_weight = pair[0].parameters[0].value
            op_second_weight = pair[-1].parameters[0].value
            op_first_max = op_first_weight.abs().max()
            op_second_max = op_second_weight.abs().max()
            scale = (op_second_max / op_first_max).sqrt()
            pair[0].parameters[0].value = pair[0].parameters[0].value * scale
            if len(pair[0].parameters) > 1:
                pair[0].parameters[1].value = pair[0].parameters[1].value * scale
            pair[-1].parameters[0].value = pair[-1].parameters[0].value / scale

    def prepare_weight_for_equalization(
        self, pair: List[Operation]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        first_computing_op_weight = pair[0].parameters[0].value
        last_computing_op_weight = pair[-1].parameters[0].value
        first_weight_range = torch.tensor(0)
        last_weight_range = torch.tensor(0)

        if pair[0].type == "Conv":
            # [C_out, C_in, K, K]
            c_out = first_computing_op_weight.shape[0]
            first_weight_range = (
                first_computing_op_weight.reshape(c_out, -1).abs().max(dim=1)[0]
            )
        elif pair[0].type == "Gemm":
            if pair[0].attributes.get("transB", 0):
                # [C_out, C_in]
                first_weight_range = first_computing_op_weight.abs().max(dim=1)[0]
            else:
                # [C_in, C_out]
                first_weight_range = first_computing_op_weight.abs().max(dim=0)[0]
        elif pair[0].type == "ConvTranspose":
            # [C_in, C_out // g, K, K]
            num_group = pair[0].attributes.get("group", 1)
            c_in, c_out_g, k1, k2 = first_computing_op_weight.shape
            first_computing_op_weight = first_computing_op_weight.reshape(
                num_group, c_in // num_group, c_out_g, k1, k2
            )
            first_computing_op_weight = first_computing_op_weight.permute(
                0, 2, 1, 3, 4
            ).contiguous()
            first_computing_op_weight = first_computing_op_weight.reshape(
                num_group * c_out_g, -1
            )
            first_weight_range = first_computing_op_weight.abs().max(dim=1)[0]
        if pair[-1].type == "Conv":
            # [C_out, C_in // g, K, K]
            num_group = pair[-1].attributes.get("group", 1)
            c_out, C_in_g, k1, k2 = last_computing_op_weight.shape
            last_computing_op_weight = last_computing_op_weight.reshape(
                num_group, c_out // num_group, C_in_g, k1, k2
            )
            last_computing_op_weight = last_computing_op_weight.permute(
                0, 2, 1, 3, 4
            ).contiguous()
            last_computing_op_weight = last_computing_op_weight.reshape(
                num_group * C_in_g, -1
            )
            last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]
        elif pair[-1].type == "Gemm":
            c_out = first_weight_range.shape[0]
            if pair[-1].attributes.get("transB", 0):
                if c_out != last_computing_op_weight.shape[1]:
                    last_computing_op_weight = last_computing_op_weight.reshape(
                        last_computing_op_weight.shape[0], c_out, -1
                    )
                    last_computing_op_weight = (
                        last_computing_op_weight.permute(1, 0, 2)
                        .contiguous()
                        .reshape(c_out, -1)
                    )
                    last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]
                else:
                    last_weight_range = last_computing_op_weight.abs().max(dim=0)[0]
            else:
                if c_out != last_computing_op_weight.shape[0]:
                    last_computing_op_weight = last_computing_op_weight.reshape(
                        c_out, -1
                    )
                    last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]
                else:
                    last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]
        elif pair[-1].type == "ConvTranspose":
            c_in = last_computing_op_weight.shape[0]
            last_weight_range = (
                last_computing_op_weight.reshape(c_in, -1).abs().max(dim=1)[0]
            )
        return first_weight_range, last_weight_range

    def write_back(self, pair: List[Operation], scale: torch.Tensor) -> None:
        first_computing_op_weight = pair[0].parameters[0].value
        last_computing_op_weight = pair[-1].parameters[0].value

        assert isinstance(first_computing_op_weight, torch.Tensor)
        assert isinstance(last_computing_op_weight, torch.Tensor)

        if pair[0].type == "Conv":
            pair[0].parameters[0].value = first_computing_op_weight * scale.reshape(
                -1, 1, 1, 1
            )
        elif pair[0].type == "Gemm":
            if pair[0].attributes.get("transB", 0):
                pair[0].parameters[0].value = first_computing_op_weight * scale.reshape(
                    -1, 1
                )
            else:
                pair[0].parameters[0].value = first_computing_op_weight * scale.reshape(
                    1, -1
                )
        elif pair[0].type == "ConvTranspose":
            num_group = pair[0].attributes.get("group", 1)
            c_in, c_out_g, k1, k2 = first_computing_op_weight.shape
            first_computing_op_weight = first_computing_op_weight.reshape(
                num_group, c_in // num_group, c_out_g, k1, k2
            )
            first_computing_op_weight = first_computing_op_weight * scale.reshape(
                num_group, 1, -1, 1, 1
            )
            pair[0].parameters[0].value = first_computing_op_weight.reshape(
                c_in, c_out_g, k1, k2
            )
        if len(pair[0].parameters) > 1:
            pair[0].parameters[1].value = pair[0].parameters[1].value * scale
        if pair[-1].type == "Conv":
            num_group = pair[-1].attributes.get("group", 1)
            c_out, c_in_g, k1, k2 = last_computing_op_weight.shape
            last_computing_op_weight = last_computing_op_weight.reshape(
                num_group, c_out // num_group, c_in_g, k1, k2
            )
            last_computing_op_weight = last_computing_op_weight / scale.reshape(
                num_group, 1, -1, 1, 1
            )
            pair[-1].parameters[0].value = last_computing_op_weight.reshape(
                c_out, c_in_g, k1, k2
            )
        elif pair[-1].type == "Gemm":
            if pair[-1].attributes.get("transB", 0):
                if scale.numel() != last_computing_op_weight.shape[1]:
                    last_computing_op_weight = last_computing_op_weight.reshape(
                        last_computing_op_weight.shape[0], scale.numel(), -1
                    )
                    last_computing_op_weight = last_computing_op_weight / scale.reshape(
                        1, -1, 1
                    )
                    pair[-1].parameters[0].value = last_computing_op_weight.reshape(
                        last_computing_op_weight.shape[0], -1
                    )
                else:
                    pair[-1].parameters[0].value = (
                        last_computing_op_weight / scale.reshape(1, -1)
                    )
            else:
                if scale.numel() != last_computing_op_weight.shape[0]:
                    last_computing_op_weight = last_computing_op_weight.reshape(
                        scale.numel(), -1, last_computing_op_weight.shape[-1]
                    )
                    last_computing_op_weight = last_computing_op_weight / scale.reshape(
                        -1, 1, 1
                    )
                    pair[-1].parameters[0].value = last_computing_op_weight.reshape(
                        -1, last_computing_op_weight.shape[-1]
                    )
                else:
                    pair[-1].parameters[0].value = (
                        last_computing_op_weight / scale.reshape(-1, 1)
                    )

        elif pair[-1].type == "ConvTranspose":
            pair[-1].parameters[0].value = last_computing_op_weight / scale.reshape(
                -1, 1, 1, 1
            )

    def one_step_equalization(
        self,
        pair: List[Operation],
        op_act_channel_range: Dict[Operation, torch.Tensor],
        algo_type: int = 2,
        ssd_min_scale: float = 8,
        ssd_max_scale: float = 2,
        dfq_min_scale: float = 0.1,
        dfq_max_scale: float = 10,
        eps: float = 1e-8,
    ):
        r"""Equalization step with scale being calculated in the way specified
        by algo_type.

        Args:
            pair (List[Operation]): a list of operations representing a equalzation pair
            op_act_channel_range (Dict[Operation, torch.Tensor]): channel-wise
                activation range of all Conv ops in the graph
            algo_type (int, optional): minor algo type. 0 represents dfq algo,
                1~3 represents ssd algo. Defaults to 2.
            ssd_min_scale (float, optional): minimum clip value of scale for ssd algo.
                Defaults to 8.
            ssd_max_scale (float, optional): maximum clip value of scale for ssd algo.
                Defaults to 2.
            dfq_min_scale (float, optional): minimum clip value of scale for dfq algo.
                Defaults to 0.1.
            dfq_max_scale (float, optional): maximum clip value of scale for dfq algo.
                Defaults to 10.
            eps (float, optional): small constant for numerical stability.
                Defaults to 1e-8.
        """
        first_weight_range, last_weight_range = self.prepare_weight_for_equalization(
            pair
        )

        if algo_type == 0:
            # dfq
            scale = torch.sqrt(last_weight_range / (first_weight_range + eps))
            scale = torch.clamp(scale, dfq_min_scale, dfq_max_scale)

        else:
            first_weight_range = torch.where(
                first_weight_range < first_weight_range.max() * self.channel_ratio,
                first_weight_range.max() * self.channel_ratio,
                first_weight_range,
            )
            last_weight_range = torch.where(
                last_weight_range < last_weight_range.max() * self.channel_ratio,
                last_weight_range.max() * self.channel_ratio,
                last_weight_range,
            )

            kernel_scale = first_weight_range.max() / (first_weight_range + eps)
            next_kernel_scale = last_weight_range.max() / (last_weight_range + eps)
            first_weight_act_range = op_act_channel_range[pair[0]]
            first_weight_act_range = torch.where(
                first_weight_act_range < 0.01,
                torch.tensor(
                    0.01, device=first_weight_act_range.device, dtype=torch.float32
                ),
                first_weight_act_range,
            )
            act_scale = first_weight_act_range.max() / (first_weight_act_range + eps)

            if algo_type == 1:
                scale = torch.min(kernel_scale, act_scale)
            elif algo_type == 2:
                kernel_scale = kernel_scale / next_kernel_scale
                act_scale = act_scale / next_kernel_scale
                scale = torch.min(kernel_scale, act_scale)
                scale = torch.min(
                    scale,
                    torch.tensor(
                        ssd_min_scale, dtype=torch.float32, device=scale.device
                    ),
                )
                scale /= scale.min()
                scale = torch.clamp(scale, 1.0, ssd_max_scale)
            else:
                kernel_scale = (kernel_scale / next_kernel_scale).sqrt()
                scale = (act_scale * kernel_scale).sqrt()
                scale = torch.clamp(scale, 1.0, ssd_max_scale)

        self.write_back(pair, scale)

    def build_observer_pair(
        self, pair: List[Operation]
    ) -> Dict[Operation, OperationObserver]:
        observers = {}
        for operation in pair:
            if not isinstance(operation, QuantableOperation):
                continue
            observer = OperationObserver(operation=operation)
            observers[operation] = observer
        return observers

    def calibrate(
        self,
        pair: List[Operation],
        data_loader: Iterable,
        executor: BaseGraphExecutor,
        hooks: Dict[Operation, CalibrationHook],
        collate_fn: Optional[Callable] = None,
        calib_steps: int = 0,
    ):
        for step, data in enumerate(cycle(data_loader)):
            if step >= calib_steps:
                break
            if collate_fn is not None:
                data = collate_fn(data)
            # get the input of first op
            inputs = executor.forward(data, output_names=[pair[0].inputs[0].name])
            self.run_pair(executor, pair, inputs, hooks)

    def calibration_passive_param(
        self, pair: List[Operation], scale_multiplier: float = 1.0
    ):
        for op in pair:
            if not isinstance(op, QuantableOperation):
                continue
            if op.type in {"Conv", "ConvTranspose", "Gemm"}:
                if op.num_of_input == 3:
                    weight_config = op.config.input_quantization_config[1]
                    input_config = op.config.input_quantization_config[0]
                    weight_config = weight_config.dominated_by
                    input_config = input_config.dominated_by

                    bias_config = op.config.input_quantization_config[-1]
                    if bias_config.state != QuantizationStates.PASSIVE_INIT:
                        continue

                    bias_config.scale = (
                        weight_config.scale * input_config.scale * scale_multiplier
                    )
                    bias_config.state = QuantizationStates.PASSIVE
                    bias_config.offset = torch.zeros_like(
                        bias_config.scale, dtype=torch.float
                    )
                    assert not bias_config.policy.has_property(
                        QuantizationProperty.ASYMMETRIC
                    ), "Negative parameter does not support ASYMMETRICAL quantization"

    def initiate_pair_state(self, pair: List[Operation]):
        for op in pair:
            if isinstance(op, QuantableOperation):
                for quant_config in (
                    op.config.input_quantization_config
                    + op.config.output_quantization_config
                ):
                    if quant_config.state == QuantizationStates.ACTIVATED:
                        quant_config.state = QuantizationStates.INITIAL
                    elif quant_config.state == QuantizationStates.PASSIVE:
                        quant_config.state = QuantizationStates.PASSIVE_INIT

    def dequantize_pair(self, pair: List[Operation]):
        for op in pair:
            if isinstance(op, QuantableOperation):
                op.dequantize(expire_device=None)

    def restore_quantize_state(self, pair: List[Operation]):
        for op in pair:
            if isinstance(op, QuantableOperation):
                op.restore_quantize_state(expire_device=None)

    def run_pair(
        self,
        executor: BaseGraphExecutor,
        pair: List[Operation],
        inputs: List[torch.Tensor],
        hooks: Dict[Operation, CalibrationHook] = {},
    ) -> List[torch.Tensor]:
        for op in pair:
            inputs = [*inputs] + [param.value for param in op.parameters]
            if isinstance(op, QuantableOperation):
                input_configs = [_ for _ in op.config.input_quantization_config]
                assert len(inputs) == len(input_configs)
                inputs_quant = [
                    self.quant_func(input, config)
                    for input, config in zip(inputs, input_configs)
                ]
                hook = hooks.get(op, None)
                if hook is not None:
                    hook.pre_forward_hook(inputs, inputs_quant, input_configs)
            else:
                inputs_quant = inputs

            outputs = executor.forward_single_operation(op, inputs_quant)

            if isinstance(op, QuantableOperation):
                output_configs = [_ for _ in op.config.output_quantization_config]
                outputs_quant = [
                    self.quant_func(output, config)
                    for output, config in zip(outputs, output_configs)
                ]
                hook = hooks.get(op, None)
                if hook is not None:
                    hook.post_forward_hook(outputs, outputs_quant, output_configs)
                inputs = outputs_quant
            else:
                inputs = list(outputs)
        return inputs

    def calculate_mse(
        self, fp_res: List[torch.Tensor], quant_res: List[torch.Tensor]
    ) -> torch.Tensor:
        """MSE calculation for a list of tensors"""
        losses = []
        for x, y in zip(fp_res, quant_res):
            losses.append(torch_mean_square_error(x, y))
        return torch.stack(losses).mean()

    @torch.no_grad()
    def test_ssd_loss(
        self,
        pair: List[Operation],
        executor: BaseGraphExecutor,
        data_loader: Iterable,
        collate_fn: Optional[Callable] = None,
        calib_steps: int = 0,
    ) -> float:
        observers = self.build_observer_pair(pair)
        hooks = {op: observers[op].hook for op in observers}
        self.calibrate(pair, data_loader, executor, hooks, collate_fn, calib_steps)
        for _, observer in observers.items():
            observer.render_quantization_config()
        pop_list = []
        for op, observer in observers.items():
            if all(
                [
                    type(var_observer) not in {OBSERVER_TABLE["kl"]}
                    for var_observer in observer.hook._observer_table.values()
                ]
            ):
                pop_list.append(op)
        for op in pop_list:
            observers.pop(op)
            hooks.pop(op)
        if len(hooks) > 0:
            self.calibrate(pair, data_loader, executor, hooks, collate_fn, calib_steps)
            for _, observer in observers.items():
                observer.render_quantization_config()
        self.calibration_passive_param(pair)
        # calculate loss
        loss = []
        for step, data in enumerate(cycle(data_loader)):
            if step >= calib_steps:
                break
            if collate_fn is not None:
                data = collate_fn(data)
            # get the input of first op
            inputs = executor.forward(data, output_names=[pair[0].inputs[0].name])
            # dequant to get fp output
            self.dequantize_pair(pair)
            fp_output = self.run_pair(executor, pair, inputs)
            # restore quant state to get quant output
            self.restore_quantize_state(pair)
            quant_output = self.run_pair(executor, pair, inputs)
            # mse calculation
            loss.append(self.calculate_mse(fp_output, quant_output))
        return torch.stack(loss).mean().item()

    def collect_original_parameter(
        self, pair: List[Operation]
    ) -> Dict[Variable, torch.Tensor]:
        """maintain original parameter for restoration in case of a larger
        loss after equalization
        """
        original_weights = {}
        for op in pair:
            for var in op.inputs + op.outputs:
                if var.is_parameter:
                    original_weights[var] = var.value.clone()
        return original_weights

    def store_parameter(self, pair: List[Operation]):
        """discard original fp value and store the current value as new fp value"""
        for op in pair:
            if isinstance(op, QuantableOperation):
                op.store_parameter_value()

    def recover_original_parameter(
        self,
        pair: List[Operation],
        original_weights: Dict[Variable, torch.Tensor],
    ):
        """recover from maintained original parameters"""
        for op in pair:
            for var in op.inputs + op.outputs:
                if var.is_parameter:
                    var.value = original_weights[var]
        self.store_parameter(pair)

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        collate_fn: Optional[Callable] = None,
        calib_steps: int = 0,
        **kwargs,
    ) -> None:
        assert executor is not None
        assert dataloader is not None
        # restrain maximum img number used for loss checking
        batchsize = 1
        data = next(iter(dataloader))
        if collate_fn is not None:
            data = collate_fn(data)
        if isinstance(data, torch.Tensor):
            batchsize = data.shape[0]
        elif isinstance(data, (list, tuple)):
            for value in data:
                if isinstance(value, torch.Tensor):
                    batchsize = value.shape[0]
                    break
        elif isinstance(data, dict):
            for value in data.values():
                if isinstance(value, torch.Tensor):
                    batchsize = value.shape[0]
                    break

        calib_steps = min(calib_steps, ceil(200 / batchsize))

        all_pairs = self.collect_all_pairs(graph)
        if self.layer_norm:
            self.layer_weight_norm(all_pairs)

        for i in range(self.iteration):
            debug(f"DFQ/SSD Equalization Iteration {i + 1}/{self.iteration}")
            for _, pair in tqdm(
                enumerate(all_pairs),
                desc=f"SSD/DFQ Equalization Iteration {i + 1}/{self.iteration}",
                total=len(all_pairs),
            ):
                debug(
                    f"Now Processing Pair {_ + 1}/{len(all_pairs)}: "
                    f"{'--'.join([op.name for op in pair])}"
                )
                self.store_parameter(pair)

                debug("Collecting Activation Range for Pair...")
                op_act_range = self.collect_activation_range(
                    pair, executor, dataloader, collate_fn, calib_steps
                )
                debug("Collecting Done!")

                original_weights = self.collect_original_parameter(pair)
                basic_loss = self.test_ssd_loss(
                    pair, executor, dataloader, collate_fn, calib_steps
                )
                best_loss = basic_loss
                best_idx = -1

                # now apply equalization and estimate loss
                for algo in range(0, 4):
                    self.one_step_equalization(pair, op_act_range, algo)
                    self.store_parameter(pair)
                    self.initiate_pair_state(pair)
                    loss = self.test_ssd_loss(
                        pair, executor, dataloader, collate_fn, calib_steps
                    )
                    if algo == 0:
                        debug(
                            f"DFQ Step, Loss Before Equalization {basic_loss} ||"
                            f" Loss After Equalization {loss}"
                        )
                    else:
                        debug(
                            f"SSD Algo {algo}, Loss Before Equalization {basic_loss} ||"
                            f" Loss After Equalization {loss}"
                        )
                    if loss < basic_loss * self.loss_threshold and loss < best_loss:
                        best_idx = algo
                        best_loss = loss
                    self.recover_original_parameter(pair, original_weights)

                if best_idx >= 0:
                    if best_idx == 0:
                        debug(
                            f"DFQ Step Activated, Loss Before Equalization {basic_loss} "
                            f"|| Loss After Equalization {best_loss}"
                        )
                    else:
                        debug(
                            f"SSD Algo {best_idx} Activated, Loss Before Equalization "
                            f"{basic_loss} || Loss After Equalization {best_loss}"
                        )
                    self.one_step_equalization(pair, op_act_range, best_idx)
                    self.store_parameter(pair)
                else:
                    debug("SSD and DFQ Deactivated")
                self.initiate_pair_state(pair)



from mppq.quantization.analyse.graphwise import (
    graphwise_error_analyse,
    statistical_analyse,
)
from mppq.quantization.analyse.layerwise import (
    layerwise_error_analyse,
    parameter_analyse,
    variable_analyse,
)

__all__ = [
    "graphwise_error_analyse",
    "statistical_analyse",
    "layerwise_error_analyse",
    "parameter_analyse",
    "variable_analyse",
]



from typing import Tuple

import torch
from torch.autograd import Function

from mppq.common import OBSERVER_MIN_SCALE, OBSERVER_MIN_SCALE_MANUL_OVERRIDE
from mppq.ffi import CUDA, ENABLE_CUDA_KERNEL
from mppq.logger import warning
from mppq.quant import QuantizationProperty, RoundingPolicy, TensorQuantizationConfig
from mppq.utils.round import (
    ppq_numerical_round,
    ppq_round_to_power_of_2,
    ppq_tensor_round,
)


def minmax_to_scale_offset(
    min_val: float,
    max_val: float,
    config: TensorQuantizationConfig,
    scale_threshold: float = OBSERVER_MIN_SCALE,
) -> Tuple[float, float]:
    """
    Solve scale and offset with given min, max value.
    For Symmetrical Quantization, offset is set to 0.
    For ASymmetrical Quantization, offset is limited by
    [config.quant_min, config.quant_max].

    Scale is limited by [scale_threshold, +inf].

    Args:
        min_val (float): min value
        max_val (float): max value
        config (TensorQuantizationConfig): Corresponding TQC.
        scale_threshold (float, optional): minimum scale.

    Returns:
        Tuple[float, float]: Solved scale and offset.
    """
    if OBSERVER_MIN_SCALE_MANUL_OVERRIDE in config.detail:
        scale_threshold = config.detail[OBSERVER_MIN_SCALE_MANUL_OVERRIDE]

    scale, offset = 1, 0
    if min_val > 0:
        min_val = 0
    if max_val < 0:
        max_val = 0

    if config.policy.has_property(QuantizationProperty.ASYMMETRIC):
        val_range = float(max_val - min_val)
        scale = val_range / (config.quant_max - config.quant_min)
        if scale < scale_threshold:
            warning(
                "Numeric instability detected: "
                "ppq find there is a scale value < 1e-7, "
                "which probably cause numeric underflow in further computation."
            )
        scale = max(scale, scale_threshold)
        offset = ppq_numerical_round(-min_val / scale)
    elif config.policy.has_property(QuantizationProperty.SYMMETRIC):
        val_range = 2 * float(max(abs(max_val), abs(min_val)))
        scale = val_range / (config.quant_max - config.quant_min)
        if scale < scale_threshold:
            warning(
                "Numeric instability detected: "
                "ppq find there is a scale value < 1e-7, "
                "which probably cause numeric underflow in further computation."
            )
        scale = max(scale, scale_threshold)
        offset = 0
    else:
        raise TypeError(
            "Tensor Min Max Observer Excepts either ASYMMETRICAL or SYMMETRICAL "
            "quantization config."
        )
    if config.policy.has_property(QuantizationProperty.POWER_OF_2):
        scale = ppq_round_to_power_of_2(scale, policy=RoundingPolicy.ROUND_UP)
    return scale, offset


# pylint: disable=abstract-method, arguments-differ
class TensorwiseLinearQuantImpl(Function):
    r"""Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        quant_min: int,
        quant_max: int,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)

        if not ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL or not tensor.is_cuda:
            # quantization function, pytorch implementation
            tensor = ppq_tensor_round((tensor / scales), rounding) + offsets
            tensor = torch.clamp(tensor, quant_min, quant_max)
            tensor = (tensor - offsets) * scales
            return tensor
        else:
            # quantization function, pure cuda implementation
            quantized = CUDA.LinearQuantize_T(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value,
            )
            return quantized

    @staticmethod
    def backward(ctx, *grad_outputs: torch.Tensor):
        return grad_outputs[0], None, None, None, None, None, None, None, None


class ChannelwiseLinearQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        quant_min: int,
        quant_max: int,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)

        if not ENABLE_CUDA_KERNEL.USING_CUDA_KERNEL or not tensor.is_cuda:
            # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
            shape = [1 if axis != channel_axis else -1 for axis in range(tensor.ndim)]
            scale, offset = scales.view(shape), offsets.view(shape)

            tensor = ppq_tensor_round((tensor / scale), rounding) + offset
            tensor = torch.clamp(tensor, quant_min, quant_max)
            tensor = (tensor - offset) * scale
            return tensor
        else:

            quantized = CUDA.LinearQuantize_C(
                tensor=tensor,
                scales=scales,
                offsets=offsets,
                channel_axis=channel_axis,
                minimum=quant_min,
                maximum=quant_max,
                rounding=rounding.value,
            )
            return quantized

    @staticmethod
    def backward(ctx, *grad_outputs: torch.Tensor):
        return grad_outputs[0], None, None, None, None, None, None, None, None, None


class TensorwiseDynamicLinearQuantImpl(Function):
    r"""Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx, tensor: torch.Tensor, config: TensorQuantizationConfig
    ) -> torch.Tensor:
        # solve scale and offset at first.
        scales, offsets = minmax_to_scale_offset(
            tensor.min().item(), tensor.max().item(), config=config
        )
        print(scales, offsets)
        # quantization function, pytorch implementation
        tensor = ppq_tensor_round((tensor / scales), config.rounding) + offsets
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
        tensor = (tensor - offsets) * scales
        return tensor

    @staticmethod
    def backward(ctx, *grad_outputs: torch.Tensor):
        return grad_outputs[0], None


class ChannelwiseDynamicLinearQuantImpl(Function):
    r"""Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx, tensor: torch.Tensor, config: TensorQuantizationConfig
    ) -> torch.Tensor:
        channelwise_view = tensor.transpose(dim0=0, dim1=config.channel_axis).unsqueeze(
            -1
        )
        channelwise_view = torch.flatten(channelwise_view, start_dim=1)

        scales, offsets = [], []
        for _min, _max in zip(
            channelwise_view.min(dim=1)[0].tolist(),
            channelwise_view.max(dim=1)[0].tolist(),
        ):
            s, o = minmax_to_scale_offset(_min, _max, config)
            scales.append(s)
            offsets.append(o)

        scales = torch.tensor(scales, dtype=torch.float32, device=tensor.device)
        offsets = torch.tensor(offsets, dtype=torch.float32, device=tensor.device)

        # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
        shape = [
            1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)
        ]
        scales, offsets = scales.view(shape), offsets.view(shape)

        tensor = ppq_tensor_round((tensor / scales), config.rounding) + offsets
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
        tensor = (tensor - offsets) * scales
        return tensor

    @staticmethod
    def backward(ctx, *grad_outputs: torch.Tensor):
        return grad_outputs[0], None


def dynamic_linear_quant(
    tensor: torch.Tensor, config: TensorQuantizationConfig
) -> torch.Tensor:
    """
    Dynamic Linear Quantization Function(PPQ 动态量化函数).

    When calling this method, we firstly solve a scale & offset setting
    by min-max observer.

    Then we applies ordinary Linear Quantization Function with solved setting.

    If there is a pre-defined scale & offset within given config,
    they will be dropped without warning.

    动态量化函数将在执行量化之前统计出 tensor 的 min - max, 而后计算出 scale & offset 并完成量化

    此时 TQC 中的 scale 与 offset 将被忽略
    """
    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        qtensor = ChannelwiseDynamicLinearQuantImpl.apply(tensor, config)
    else:
        qtensor = TensorwiseDynamicLinearQuantImpl.apply(tensor, config)
    assert isinstance(qtensor, torch.Tensor)
    return qtensor


def linear_fake_quant(
    tensor: torch.Tensor, config: TensorQuantizationConfig
) -> torch.Tensor:
    """PPQ 核心量化函数，没啥好说的了吧，这个玩意既做 quant 也做 dequant"""
    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        qtensor = ChannelwiseLinearQuantImpl.apply(
            tensor,
            config.scale,
            config.offset,
            config.channel_axis,
            config.quant_min,
            config.quant_max,
            config.rounding,
        )
    else:
        qtensor = TensorwiseLinearQuantImpl.apply(
            tensor,
            config.scale,
            config.offset,
            config.quant_min,
            config.quant_max,
            config.rounding,
        )
    assert isinstance(qtensor, torch.Tensor)
    return qtensor


def linear_quant_toint(
    tensor: torch.Tensor, config: TensorQuantizationConfig
) -> torch.Tensor:
    """PPQ 核心量化函数，没啥好说的了吧，这个玩意只做 quant 不做 dequant"""
    if not config.policy.has_property(QuantizationProperty.LINEAR):
        raise ValueError("Critical Quantization Error! Non-linear config detected.")
    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        shape = [
            1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)
        ]
        scale, offset = config.scale.view(shape), config.offset.view(shape)
        tensor = ppq_tensor_round((tensor / scale), config.rounding) + offset
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
    else:  # config.policy.has_property(QuantizationProperty.PER_TENSOR)
        tensor = (
            ppq_tensor_round((tensor / config.scale), config.rounding) + config.offset
        )
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
    if config.num_of_bits <= 8:
        if config.num_of_bits < 8:
            warning("bits < 8 is still experimental")
        if config.policy.has_property(QuantizationProperty.SYMMETRIC):
            return tensor.type(dtype=torch.int8)
        else:  # config.policy.has_property(QuantizationProperty.ASYMMETRIC)
            return tensor.type(dtype=torch.uint8)
    else:  # config.num_of_bits > 8
        return tensor.type(dtype=torch.int32)



"""
This File Contains Legacy Dispatchers.
Refer to ppq.scheduler.perseus for updated implementation.
"""

from typing import Collection, Dict

from mppq.dispatcher.base import (
    DISPATCHER_TABLE,
    GraphDispatcher,
    reverse_tracing_pattern,
    soi_generators,
    soi_receivers,
    value_tracing_pattern,
)
from mppq.ir.base.graph import BaseGraph
from mppq.ir.search import SearchableGraph
from mppq.quant import TargetPrecision


@DISPATCHER_TABLE.register("aggressive")
class AggressiveDispatcher(GraphDispatcher):
    r"""Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable
    operations, Shape-Or-Index related operations and remaining operations, all sub
    classes of GraphDispatcher will give an implementation of function "dispatch" to
    send all operations to their proper platform.

    ATTENTION:

        platform attribute will greatly affect quantizer's quantization logic,
        and the execution result.

        - If operation is sent to a quantable platform, then its inputs and outputs will
          be quantized if necessary.
        - if operation is classified as shape-or-index related operation, then its
          execution will be taken with cpu.
        - if operation is sent to a fp32 platform, then its inputs and outputs shall
          never be quantized.

    ATTENTION:

        this dispatcher will insert necessary DeviceSwitch operations between
        shape-or-index operations and others.
    """

    def __init__(self, graph: BaseGraph) -> None:
        self.graph = graph

    def dispatch(
        self,
        quant_types: Collection[str],
        quant_precision: TargetPrecision,
        fp32_platform: TargetPrecision = TargetPrecision.FP32,
        soi_platform: TargetPrecision = TargetPrecision.SOI,
        **kwargs,
    ) -> Dict[str, TargetPrecision]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.

        There are 3 default platform during dispatching:
            quant_precision - all quantable parts of graph will be dispatched to this
                              platform
            SOI_platform   - Aka. Shape or Index related operations will be dispatched
                             to this platform.
            fp32_platform  - there are some operations receiving results from both
                             quant_precision and SOI_platform, they will be dispatched
                             they will be dispatched to fp32_platform.

        ATTENTION:

            Quantization follows this dispatching, and only the operations within
            quantable platform will be quantized in the future.

        ATTENTION:

            this dispatcher will insert necessary DeviceSwitch operations between
            shape-or-index operations and others.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this
                dispatcher.
            quant_types(Set[str]): all quantable types for given platforms.
            quant_precision (TargetPlatform):
                platform object where quantable parts will goes to.
            SOI_platform (TargetPlatform):
                platform object where SOI parts will goes to.
            fp32_platform (TargetPlatform):
                platform object where remaining parts will goes to.
        """
        graph = self.graph
        receivers, generators = soi_receivers(graph), soi_generators(graph)
        search_engine, soi_operations = (SearchableGraph(graph), set(receivers))

        quant_operations = search_engine.opset_matching(
            sp_expr=lambda op: op.is_computing_op,
            rp_expr=value_tracing_pattern,
            ep_expr=lambda op: op.type in {"Shape", "TopK", "NonMaxSuppression"}
            or op.is_boundary,
            direction="down",
        )
        # remove shape operations from computing ops.
        quant_operations.filter(lambda op: op.type == "Shape")

        # we assume all 'Shape', 'NonMaxSuppression', 'ConstantOfShape', 'Topk'
        # operations are SOI generators.
        shape_forward_matching = search_engine.opset_matching(
            sp_expr=lambda op: op in generators and op.type not in {"Constant"},
            rp_expr=value_tracing_pattern,
            ep_expr=lambda op: (
                op in receivers or op in quant_operations or op.is_boundary
            ),
            direction="down",
        )

        # update matchings, ready for further searching.
        soi_operations.update(shape_forward_matching)
        generators.update(soi_operations)

        while True:
            # there are some particular cases where a single matching can not handle.
            # to cover all shape-related operations, a reverse matching is required.
            shape_backward_matching = search_engine.opset_matching(
                sp_expr=lambda op: (
                    op in soi_operations
                    and op.type != "Shape"
                    and op not in quant_operations
                ),
                rp_expr=reverse_tracing_pattern,
                ep_expr=lambda op: (
                    op in generators or op in quant_operations or op.is_boundary
                ),
                direction="up",
            )

            if all([(op in soi_operations) for op in shape_backward_matching]):
                break
            # update matchings
            soi_operations.update(shape_backward_matching)

        # generate dispatching table.
        dispatching_table = {}
        for operation in graph.operations.values():
            if operation in soi_operations:
                dispatching_table[operation.name] = soi_platform
            elif operation in quant_operations:
                dispatching_table[operation.name] = quant_precision
            else:
                dispatching_table[operation.name] = fp32_platform

            # move Topk, Shape, NonMaxSuppression to platform as same as their input.
            if operation.type in {"Shape", "TopK", "NonMaxSuppression"}:
                if operation.inputs[0].source_op is not None:
                    dispatching_table[operation.name] = operation.inputs[
                        0
                    ].source_op.precision
                else:
                    dispatching_table[operation.name] = quant_precision

            # move activations to the platform same as their input.
            if operation.is_linear_activation:
                source_op = operation.inputs[0].source_op
                if source_op is not None:
                    dispatching_table[operation.name] = dispatching_table[
                        source_op.name
                    ]

        return dispatching_table



from collections import defaultdict
from itertools import cycle
from typing import Callable, Dict, Iterator, List, Optional, Union

import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from mppq.common import PASSIVE_OPERATIONS
from mppq.executor import TorchExecutor
from mppq.executor.base import BaseGraphExecutor
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.opdef import Operation, Variable
from mppq.ir.base.quantize import QuantableOperation
from mppq.ir.deploy import QuantableGraph
from mppq.logger import warning
from mppq.quantization.analyse.util import (
    DetailedRecorder,
    MeasurePrinter,
    MeasureRecorder,
    OutputRecorder,
)
from mppq.quantization.measure.norm import torch_snr_error


def graphwise_error_analyse(  # noqa: C901, too complex
    executor: BaseGraphExecutor,
    dataloader: DataLoader,
    interested_outputs: Optional[Union[str, List[str]]] = None,
    collate_fn: Optional[Callable] = None,
    method: str = "snr",
    reduce: str = "mean",
    steps: int = 8,
    verbose: bool = False,
    fetches: int = 4096,
) -> Dict[str, float]:
    """Measure the difference from a quantized graph to its dequantized graph.

    A dictionary contains output differences for all operation will be returned as a
    result.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}

    if verbose is set as True, this function will display error report at last.

    The key of the dictionary is an operation name while the value of corresponding key
    is the difference between quantized output and float output of this operation.

    Result {'operation name 1': 0.933} means that quantized graph and fp32 graph have a
    difference (or similarity, based on your measurement) 0.933 at the output tensor of
    'operation name 1'.

    ATTENTION:

        Output difference is measured at graph-level, it includes the difference
        accmulated from the very beginning operation along to the target operation.

    Args:
        graph (BaseGraph):
            A fully quantized graph instance.

        dataloader (Iterator):
            Test dataloader, this function will measure the output difference based on
            given data.

        interested_outputs (Union[str, List[str]] = None)
            a list contains your interested output variables. if set as None, all graph
            output variables will be measured via this function.

        collate_fn (Callable, optional):
            An data preprocessing function provided by user to convert data from
            dataloader towards executable format. If set as None, then no action will
            be taken during preprocessing.

        method (str, optional):
            A string indicates a measurement to calculate the difference of quantized
            output and fp32 one. 'cosine', 'snr', and 'mse' is supported in PPQ for now.

        steps (Int, optional)
            computation steps.

    Returns:
        A dictionary contains output differences for all operation will be returned
        from this function.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}
    """
    graph = executor._graph  # pylint: disable=protected-access
    assert graph is not None
    quant_helper = QuantableGraph(graph)

    # find all quantable operations.
    interested_op = [
        operation
        for operation in graph.operations.values()
        if (isinstance(operation, QuantableOperation) and operation.is_computing_op)
    ]
    if len(interested_op) == 0:
        warning("Oops. you got nothing to analyse.")
        return {}
    if interested_outputs is None:
        interested_outputs = [name for name in graph.outputs]
    if isinstance(interested_outputs, str):
        interested_outputs = [interested_outputs]

    # set up all hooks.
    hooks: Dict[str, OutputRecorder] = {}
    caches: Dict[str, List[torch.Tensor]] = defaultdict(list)
    recorders: Dict[str, MeasureRecorder] = {}
    for operation in interested_op:
        if isinstance(operation, QuantableOperation):
            if operation.num_of_output > 1:
                warning(
                    f"Operation {operation.name} has more than 1 output, "
                    "analyser will process the first output of it."
                )

            recorders[operation.name] = MeasureRecorder(
                measurement=method, reduce=reduce
            )
            hooks[operation.name] = OutputRecorder(operation=operation, fetches=fetches)
    for output in interested_outputs:
        if output in recorders:
            raise ValueError(f"output {output} has name confliction with a node!")
        recorders[output] = MeasureRecorder(measurement=method, reduce=reduce)

    # dequantize all
    quant_helper.dequantize_graph()

    # run for each quantable operations:
    for idx, batch in tqdm(
        enumerate(dataloader),
        desc="Analysing Graphwise Quantization Error(Phrase 1):",
        total=(min(len(dataloader), steps)),
    ):
        if collate_fn is not None:
            batch = collate_fn(batch)
        fp_outputs = executor.forward(
            inputs=batch, hooks=hooks, output_names=interested_outputs
        )

        for operation in interested_op:
            hook = hooks[operation.name]
            caches[operation.name].append(hook.pop())
        for key, value in zip(interested_outputs, fp_outputs):
            caches[key].append(value)

        if idx >= steps:
            break

    # restore all
    quant_helper.restore_quantize_state()

    # run for each quantable operations:
    for idx, batch in tqdm(
        enumerate(dataloader),
        desc="Analysing Graphwise Quantization Error(Phrase 2):",
        total=(min(len(dataloader), steps)),
    ):
        if collate_fn is not None:
            batch = collate_fn(batch)
        qt_outputs = executor.forward(
            inputs=batch, hooks=hooks, output_names=interested_outputs
        )

        for operation in interested_op:
            recorder = recorders[operation.name]
            hook = hooks[operation.name]
            cache = caches[operation.name]
            recorder.update(y_real=cache[idx], y_pred=hook.pop())
        for key, value in zip(interested_outputs, qt_outputs):
            recorder = recorders[key]
            cache = caches[key]
            recorder.update(y_real=cache[idx], y_pred=value)

        if idx >= steps:
            break

    results = {name: recorder.measure for name, recorder in recorders.items()}

    if verbose:
        method_str = "MEASUREMENT"
        if method == "snr":
            method_str = "NOISE:SIGNAL POWER RATIO"
        if method == "cosine":
            method_str = "COSINE SIMILARITY"
        if method == "mse":
            method_str = f"{reduce.upper()} SQUARE ERROR"
        MeasurePrinter(
            results,
            order="large_to_small",
            measure=method_str,
            percentage=method in {"snr", "cosine"},
        ).print()
    return results


def statistical_analyse(  # noqa: C901
    graph: BaseGraph,
    running_device: str,
    dataloader: Iterator,
    collate_fn: Optional[Callable] = None,
    steps: int = 8,
) -> List[dict]:
    """It is time to do some statistical work.

    statistical_analyse is a powerful analyzing function
        that provides a in-depth study of your network.

    use report = statistical_analyse() to invoke this function

    The return value of this function is a collection of statistics parameters
    You are recommended to processing them with pandas

    from pandas import DataFrame
    report_df = DataFrame(report)

    Args:
        graph (BaseGraph): _description_
        running_device (str): _description_
        dataloader (Iterator): _description_
        collate_fn (Callable, optional): _description_. Defaults to None.
        steps (int, optional): _description_. Defaults to 8.

    Returns:
        Dict[str, float]: _description_
    """

    class StatisticalErrorAnalyser:
        def __init__(
            self,
            x_fp: List[torch.Tensor],
            x_qt: List[torch.Tensor],
            op: Operation,
            var: Variable,
        ) -> None:
            self.x_qt = torch.cat(x_qt, dim=0)
            self.x_fp = torch.cat(x_fp, dim=0)
            self.x_er = self.x_qt - self.x_fp
            self.op = op
            self.var = var

            self.num_of_samples = self.x_fp.shape[0]

        def stat(self) -> dict:
            x_er, x_fp, x_qt = self.x_er, self.x_fp, self.x_qt
            er_mean = x_er.mean().item()
            er_std = x_er.std().item()
            er_min = x_er.min().item()
            er_max = x_er.max().item()
            er_skew = self.solve_skewness(x_er, er_mean, er_std).item()
            er_kurtosis = self.solve_kurtosis(x_er, er_mean, er_std).item()
            er_hist = (
                torch.histc(x_er, bins=32, min=x_er.min().item(), max=x_er.max().item())
                .cpu()
                .tolist()
            )

            qt_mean = x_qt.mean().item()
            qt_std = x_qt.std().item()
            qt_min = x_qt.min().item()
            qt_max = x_qt.max().item()
            qt_skew = self.solve_skewness(x_qt, qt_mean, qt_std).item()
            qt_kurtosis = self.solve_kurtosis(x_qt, qt_mean, qt_std).item()
            qt_hist = (
                torch.histc(x_qt, bins=32, min=x_qt.min().item(), max=x_qt.max().item())
                .cpu()
                .tolist()
            )

            fp_mean = x_fp.mean().item()
            fp_std = x_fp.std().item()
            fp_min = x_fp.min().item()
            fp_max = x_fp.max().item()
            fp_skew = self.solve_skewness(x_fp, fp_mean, fp_std).item()
            fp_kurtosis = self.solve_kurtosis(x_fp, fp_mean, fp_std).item()
            fp_hist = (
                torch.histc(x_fp, bins=32, min=x_fp.min().item(), max=x_fp.max().item())
                .cpu()
                .tolist()
            )

            snr = torch_snr_error(x_qt, x_fp).item()
            return {
                "Op name": self.op.name,
                "Op type": self.op.type,
                "Is parameter": self.var.is_parameter,
                "Is input": self.var in self.op.inputs,
                "Is output": self.var in self.op.outputs,
                "Variable name": self.var.name,
                "Noise:Signal Power Ratio": snr,
                "Noise Mean": er_mean,
                "Noise Std": er_std,
                "Noise Skewness": er_skew,
                "Noise Kurtosis": er_kurtosis,
                "Noise Hist": er_hist,
                "Noise Max": er_max,
                "Noise Min": er_min,
                "Quantized Mean": qt_mean,
                "Quantized Std": qt_std,
                "Quantized Skewness": qt_skew,
                "Quantized Kurtosis": qt_kurtosis,
                "Quantized Hist": qt_hist,
                "Quantized Max": qt_max,
                "Quantized Min": qt_min,
                "Float Mean": fp_mean,
                "Float Std": fp_std,
                "Float Skewness": fp_skew,
                "Float Kurtosis": fp_kurtosis,
                "Float Hist": fp_hist,
                "Float Max": fp_max,
                "Float Min": fp_min,
            }

        def solve_skewness(
            self, x: torch.Tensor, mean: float, std: float
        ) -> torch.Tensor:
            # 三次方可能有数据溢出
            return torch.pow((x - mean) / std, 3).mean()

        def solve_kurtosis(
            self, x: torch.Tensor, mean: float, std: float
        ) -> torch.Tensor:
            # 四次方可能有数据溢出
            return torch.pow((x - mean) / std, 4).mean() - 3

    executor = TorchExecutor(graph=graph, device=running_device)
    # find all quantable operations.
    interested_op = []
    for operation in graph.operations.values():
        if (
            isinstance(operation, QuantableOperation)
            and operation.type not in PASSIVE_OPERATIONS
        ):
            interested_op.append(operation)
    if len(interested_op) == 0:
        print("Oops. you got nothing to analyse.")
        return []

    # set up all hooks.
    hooks, caches = {}, {}
    for operation in interested_op:
        if isinstance(operation, QuantableOperation):
            hooks[operation.name] = DetailedRecorder(operation=operation)
            caches[operation.name] = {
                "Quantized Input": [],
                "Quantized Output": [],
                "Dequantized Input": [],
                "Dequantized Output": [],
            }

    # dequantize all
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.dequantize()

    # run for each quantable operations:
    for idx, batch in tqdm(
        enumerate(cycle(dataloader)),
        desc="Analysing Phrase 1",
        total=steps,
    ):
        if collate_fn is not None:
            batch = collate_fn(batch)
        executor.forward(inputs=batch, hooks=hooks)
        if idx >= steps:
            break

    for operation in interested_op:
        hook = hooks[operation.name]
        assert isinstance(hook, DetailedRecorder)
        caches[operation.name]["Dequantized Input"] = hook.i_storage.copy()
        caches[operation.name]["Dequantized Output"] = hook.o_storage.copy()
        hook.clear()

    # restore all
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.restore_quantize_state()

    # run for each quantable operations:
    for idx, batch in tqdm(
        enumerate(cycle(dataloader)),
        desc="Analysing Phrase 2",
        total=steps,
    ):
        if collate_fn is not None:
            batch = collate_fn(batch)
        executor.forward(inputs=batch, hooks=hooks)
        if idx >= steps:
            break

    for operation in interested_op:
        hook = hooks[operation.name]
        assert isinstance(hook, DetailedRecorder)
        caches[operation.name]["Quantized Input"] = hook.i_storage.copy()
        caches[operation.name]["Quantized Output"] = hook.o_storage.copy()
        hook.clear()

    # analysing cache
    records = []
    for name, record in caches.items():
        operation = graph.operations[name]
        assert isinstance(operation, Operation)
        for idx, input_var in enumerate(operation.inputs):
            x_qt = record["Quantized Input"][idx]
            x_fp = record["Dequantized Input"][idx]
            records.append(
                StatisticalErrorAnalyser(
                    x_fp=x_fp, x_qt=x_qt, op=operation, var=input_var
                ).stat()
            )

        for idx, output_var in enumerate(operation.outputs):
            x_qt = record["Quantized Output"][idx]
            x_fp = record["Dequantized Output"][idx]
            records.append(
                StatisticalErrorAnalyser(
                    x_fp=x_fp, x_qt=x_qt, op=operation, var=output_var
                ).stat()
            )

    return records



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

"""

from typing import Type

from mppq.ir.base.graph import GraphBuilder, GraphExporter
from mppq.register import Registry

PARSER: Registry[Type[GraphBuilder]] = Registry("PARSER")
EXPORTER: Registry[Type[GraphExporter]] = Registry("EXPORTER")



from typing import Any, List, Optional

import torch

from mppq.data import convert_any_to_tensor
from mppq.ir.base.command import (
    GraphCommand,
    GraphCommandType,
    GraphDeployCommand,
    QuantizeOperationCommand,
    ReplaceOperationCommand,
    ReplaceVariableCommand,
)
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.opdef import Operation, Variable
from mppq.ir.base.processor import GraphCommandProcessor
from mppq.ir.base.quantize import QuantableOperation, QuantableVariable
from mppq.quant import OperationQuantizationConfig, TargetPrecision


class RunnableGraph(GraphCommandProcessor):
    r"""RunnableGraph deals with values related with graph executing.

    Literally it helps you move values of your graph towards device and vice versa.
        And give an executable order of all operations in your graph which actual
        executor will follow.

    Args:
        graph (BaseGraph): BaseGraph instance.
        device (str, optional): This attribute is only used by with
            RunnableGraph(graph, device) syntactic.
    """

    def __init__(self, graph: BaseGraph, device: Optional[str] = None) -> None:
        super().__init__(graph_or_processor=graph)
        self._device = device or "cpu"

    def process(self, command: GraphCommand):
        if command.command_type == GraphCommandType.DEPLOY_TO_CPU:
            return self.deploy("cpu")
        elif command.command_type == GraphCommandType.DEPLOY_TO_CUDA:
            if isinstance(command, GraphDeployCommand):
                device = command.device
                return self.deploy(device)
            else:
                return self.deploy("cuda")
        raise RuntimeError

    @property
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [GraphCommandType.DEPLOY_TO_CPU, GraphCommandType.DEPLOY_TO_CUDA]

    def deploy(self, device: str):
        for _, operator in self._graph.operations.items():
            assert isinstance(operator, Operation)
            # in onnx format, some constant values are wrapped with operation's
            # attributes['value']. To move those constant value from numpy to device,
            # we have to move all the attributes['value'] of operation to device.
            if (
                operator.type == "Constant"
                and operator.precision != TargetPrecision.SOI
            ):
                operator.attributes["value"] = convert_any_to_tensor(
                    operator.attributes["value"], device=device
                )
            if (
                operator.type == "Constant"
                and operator.precision == TargetPrecision.SOI
            ):
                operator.attributes["value"] = convert_any_to_tensor(
                    operator.attributes["value"], device="cpu"
                )
            # PATCH 20220706, send quantization config to device.
            if isinstance(operator, QuantableOperation):
                for cfg, _ in operator.config_with_variable:
                    # pylint: disable=protected-access
                    if isinstance(cfg._scale, torch.Tensor):
                        cfg._scale = cfg._scale.to(device)
                    if isinstance(cfg._offset, torch.Tensor):
                        cfg._offset = cfg._offset.to(device)

        for _, variable in self._graph.variables.items():
            assert isinstance(variable, Variable)
            # graph output variable has no destinations
            if len(variable.dest_ops) == 0:
                continue
            if not variable.has_value:
                continue

            # check all destination operations platform are same.
            precisions = [op.precision for op in variable.dest_ops]
            if (
                all([_ == precisions[0] for _ in precisions])
                and precisions[0] == TargetPrecision.SOI
            ):
                precision = TargetPrecision.SOI
            else:
                precision = TargetPrecision.UNSPECIFIED

            # if all downstream operations are shape related operations,
            # send value to cpu
            if precision == TargetPrecision.SOI:
                variable.value = convert_any_to_tensor(variable.value).to("cpu")
            else:
                variable.value = convert_any_to_tensor(variable.value).to(device=device)

            # if variable is a shape-related variable, send it to cpu.
            if variable.is_parameter:
                if len(variable.dest_ops) > 1:
                    raise PermissionError(
                        f"PPQ can not process parameter variable({variable.name})"
                        " with multiple destinations"
                        f"({[op.name for op in variable.dest_ops]}), split it first."
                    )
                dest_op = variable.dest_ops[0]
                dest_idx = dest_op.inputs.index(variable)

                assert isinstance(dest_op, Operation)
                socket = dest_op.socket
                if socket.in_plat[dest_idx] == TargetPrecision.SOI:
                    variable.value = convert_any_to_tensor(variable.value, device="cpu")
        return self


class QuantableGraph(GraphCommandProcessor):
    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.QUANTIZE_OPERATION:
            assert isinstance(command, QuantizeOperationCommand)
            return self.quantize_operation(
                command.op_name, command.target_precision, command.config
            )

    @property
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [GraphCommandType.QUANTIZE_OPERATION]

    def quantize_operation(
        self,
        operation_name: str,
        target_precision: TargetPrecision,
        quantization_config: OperationQuantizationConfig,
    ) -> QuantableOperation:
        if operation_name not in self.graph.operations:
            raise KeyError(
                f"Operation {operation_name} is not in your graph, "
                "Please check your input."
            )

        operation = self._graph.operations[operation_name]
        quantized_operation = QuantableOperation(
            convert_from=operation,
            quantize_config=quantization_config,
            platform=target_precision,
        )

        # calling other chain responder to replace operation with quantized one.
        if self._next_command_processor is None:
            raise RuntimeError(
                "To replace a operation, your processor chain must have "
                "a GraphReplacer Processor."
            )
        self._next_command_processor(
            ReplaceOperationCommand(operation_name, quantized_operation)
        )

        # replace all related variable with quantable one.
        for var in quantized_operation.inputs + quantized_operation.outputs:
            if isinstance(var, QuantableVariable):
                continue
            self._next_command_processor(
                ReplaceVariableCommand(
                    var_name=var.name, replace_to=QuantableVariable(convert_from=var)
                )
            )
        quantized_operation.store_parameter_value()
        return quantized_operation

    def dequantize_operation(self, operation_name: str) -> Operation:
        if operation_name not in self.graph.operations:
            raise KeyError(
                f"Operation {operation_name} is not in your graph, "
                "Please check your input."
            )
        operation = self._graph.operations[operation_name]
        if not isinstance(operation, QuantableOperation):
            return operation
        else:
            return operation.dequantize()

    def dequantize_graph(self, expire_device: str = "cpu"):
        """一个方便懒人的函数."""
        for operation in self.graph.operations.values():
            if isinstance(operation, QuantableOperation):
                operation.dequantize(expire_device=expire_device)

    def restore_quantize_state(self, expire_device: str = "cpu"):
        """一个方便懒人的函数."""
        for operation in self.graph.operations.values():
            if isinstance(operation, QuantableOperation):
                operation.restore_quantize_state(expire_device=expire_device)
